# Copyright 2015 Jim Unroe <rock.unroe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
import logging

from chirp import bitwise
from chirp import chirp_common
from chirp import directory
from chirp import errors
from chirp import memmap
from chirp import util
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueInteger, RadioSettingValueList, \
    RadioSettingValueBoolean, RadioSettingValueString, \
    RadioSettings

LOG = logging.getLogger(__name__)

mem_format = """
struct memory {
  bbcd freq[4];
  bbcd offset[4];
  u8 unknown1:4,
     tune_step:4;
  u8 unknown2:2,
     txdcsextra:1,
     txinv:1,
     channel_width:2,
     unknown3:1,
     tx_off:1;
  u8 unknown4:2,
     rxdcsextra:1,
     rxinv:1,
     power:2,
     duplex:2;
  u8 unknown5:4,
     rxtmode:2,
     txtmode:2;
  u8 unknown6:2,
     txtone:6;
  u8 unknown7:2,
     rxtone:6;
  u8 txcode;
  u8 rxcode;
  u8 unknown8[3];
  char name[6];
  u8 squelch:4,
     unknown9:2,
     bcl:2;
  u8 unknownA;
  u8 unknownB:7,
     sqlmode:1;
  u8 unknownC[4];
};

#seekto 0x0010;
struct {
    u8 unknown1;
    u8 unknown2:5,
       bands1:3;
    char model[7];
    u8 unknown3:5,
       bands2:3;
    u8 unknown4[6];
    u8 unknown5[16];
    char date[9];
    u8 unknown6[7];
    u8 unknown7[16];
    u8 unknown8[16];
    char dealer[16];
    char stockdate[9];
    u8 unknown9[7];
    char selldate[9];
    u8 unknownA[7];
    char seller[16];
} oem_info;

#seekto 0x0100;
u8 used_flags[50];

#seekto 0x0120;
u8 skip_flags[50];

#seekto 0x0220;
struct {
  u8 unknown220:6,
     display:2;
  u8 unknown221:7,
     upvfomr:1;
  u8 unknown222:7,
     dnvfomr:1;
  u8 unknown223:7,
     fmvfomr:1;
  u8 upmrch;
  u8 dnmrch;
  u8 unknown226:1,
     fmmrch:7;
  u8 unknown227;
  u8 unknown228:7,
     fastscano:1;       // obltr-8r only
  u8 unknown229:6,
     pause:2;
  u8 unknown22A:5,
     stop:3;
  u8 unknown22B:6,
     backlight:2;
  u8 unknown22C:6,
     color:2;
  u8 unknown22D:6,
     vdisplay:2;
  u8 unknown22E;
  u8 unknown22F:5,
     pf1key:3;
  u8 beep:1,
     alarmoff:1,
     main:1,
     radio:1,
     unknown230:1,
     allband:1,
     elimtail:1,
     monikey:1;
  u8 fastscan:1,        // termn-8r only
     keylock:1,
     unknown231:2,
     lwenable:1,
     swenable:1,
     fmenable:1,
     amenable:1;
  u8 unknown232:3,
     tot:5;
  u8 unknown233:7,
     amvfomr:1;
  u8 unknown234:3,
     apo:5;
  u8 unknown235:5,
     pf2key:3;          // keylock for obltr-8r
  u8 unknown236;
  u8 unknown237:4,
     save:4;
  u8 unknown238:5,
     tbst:3;
  u8 unknown239:4,
     voxlevel:4;
  u8 unknown23A:3,
     voxdelay:5;
  u8 unknown23B:5,
     tail:3;
  u8 unknown23C;
  u8 unknown23D:1,
     ammrch:7;
  u8 unknown23E:5,
     vvolume:3;
  u8 unknown23F:5,
     fmam:3;
  u8 unknown240:4,
     upmrbank:4;
  u8 unknown241:7,
     upwork:1;
  u8 unknown242:7,
     uplink:1;
  u8 unknown243:4,
     dnmrbank:4;
  u8 unknown244:7,
     dnwork:1;
  u8 unknown245:7,
     downlink:1;
  u8 unknown246:7,
     banklink1:1;
  u8 unknown247:7,
     banklink2:1;
  u8 unknown248:7,
     banklink3:1;
  u8 unknown249:7,
     banklink4:1;
  u8 unknown24A:7,
     banklink5:1;
  u8 unknown24B:7,
     banklink6:1;
  u8 unknown24C:7,
     banklink7:1;
  u8 unknown24D:7,
     banklink8:1;
  u8 unknown24E:7,
     banklink9:1;
  u8 unknown24F:7,
     banklink0:1;
  u8 unknown250:6,
     noaa:2;
  u8 unknown251:5,
     noaach:3;
  u8 unknown252:6,
     part95:2;
  u8 unknown253:3,
     gmrs:5;
  u8 unknown254:5,
     murs:3;
  u8 unknown255:5,
     amsql:3;
} settings;

#seekto 0x0246;
struct {
  u8 unused:7,
     bank:1;
} banklink[10];

#seekto 0x03E0;
struct {
  char line1[6];
  char line2[6];
} welcome_msg;

#seekto 0x2000;
struct memory memory[200];
"""


def _echo_write(radio, data):
    try:
        radio.pipe.write(data)
    except Exception as e:
        LOG.error("Error writing to radio: %s" % e)
        raise errors.RadioError("Unable to write to radio")


def _read(radio, length):
    try:
        data = radio.pipe.read(length)
    except Exception as e:
        LOG.error("Error reading from radio: %s" % e)
        raise errors.RadioError("Unable to read from radio")

    if len(data) != length:
        LOG.error("Short read from radio (%i, expected %i)" %
                  (len(data), length))
        LOG.debug(util.hexprint(data))
        raise errors.RadioError("Short read from radio")
    return data


valid_model = [b'TERMN8R', b'OBLTR8R']


def _ident(radio):
    radio.pipe.timeout = 1
    _echo_write(radio, b"PROGRAM")
    response = radio.pipe.read(3)
    if response != b"QX\x06":
        LOG.debug("Response was:\n%s" % util.hexprint(response))
        raise errors.RadioError("Radio did not respond. Check connection.")
    _echo_write(radio, b"\x02")
    response = radio.pipe.read(16)
    LOG.debug(util.hexprint(response))
    if radio._file_ident not in response:
        LOG.debug("Response was:\n%s" % util.hexprint(response))
        raise errors.RadioError("Unsupported model")


def _finish(radio):
    endframe = b"\x45\x4E\x44"
    _echo_write(radio, endframe)
    result = radio.pipe.read(1)
    if result != b"\x06":
        LOG.debug("Got:\n%s" % util.hexprint(result))
        raise errors.RadioError("Radio did not finish cleanly")


def _checksum(data):
    cs = 0
    for byte in data:
        cs += byte
    return cs % 256


def _send(radio, cmd, addr, length, data=None):
    frame = struct.pack(">cHb", cmd, addr, length)
    if data:
        frame += data
        frame += bytes([_checksum(frame[1:])])
        frame += b"\x06"
    _echo_write(radio, frame)
    LOG.debug("Sent:\n%s" % util.hexprint(frame))
    if data:
        result = radio.pipe.read(1)
        if result != b"\x06":
            LOG.debug("Ack was: %s" % repr(result))
            raise errors.RadioError(
                "Radio did not accept block at %04x" % addr)
        return
    result = _read(radio, length + 6)
    LOG.debug("Got:\n%s" % util.hexprint(result))
    header = result[:4]
    data = result[4:-2]
    ack = result[-1:]
    if ack != b"\x06":
        LOG.debug("Ack was: %s" % repr(ack))
        raise errors.RadioError("Radio NAK'd block at %04x" % addr)
    _cmd, _addr, _length = struct.unpack(">cHb", header)
    if _addr != addr or _length != _length:
        LOG.debug("Expected/Received:")
        LOG.debug(" Length: %02x/%02x" % (length, _length))
        LOG.debug(" Addr: %04x/%04x" % (addr, _addr))
        raise errors.RadioError("Radio send unexpected block")
    cs = _checksum(result[1:-2])
    if cs != result[-2]:
        LOG.debug("Calculated: %02x" % cs)
        LOG.debug("Actual:     %02x" % result[-2])
        raise errors.RadioError("Block at 0x%04x failed checksum" % addr)
    return data


def _download(radio):
    _ident(radio)

    memobj = None

    data = b""
    for start, end in radio._ranges:
        for addr in range(start, end, 0x10):
            block = _send(radio, b'R', addr, 0x10)
            data += block

            status = chirp_common.Status()
            status.cur = len(data)
            status.max = end
            status.msg = "Cloning from radio"
            radio.status_fn(status)

    _finish(radio)

    return memmap.MemoryMapBytes(data)


def _upload(radio):
    _ident(radio)

    for start, end in radio._ranges:
        for addr in range(start, end, 0x10):
            if addr < 0x0100:
                continue
            block = radio._mmap[addr:addr + 0x10]
            _send(radio, b'W', addr, len(block), block)

            status = chirp_common.Status()
            status.cur = addr
            status.max = end
            status.msg = "Cloning to radio"
            radio.status_fn(status)

    _finish(radio)


APO = ['Off', '30 Minutes', '1 Hour', '2 Hours']
BACKLIGHT = ['Off', 'On', 'Auto']
BCLO = ['Off', 'Repeater', 'Busy']
CHARSET = chirp_common.CHARSET_ASCII
COLOR = ['Blue', 'Orange', 'Purple']
DISPLAY = ['Frequency', 'N/A', 'Name']
DUPLEXES = ['', '', '-', '+', 'split', 'off']
GMRS = ['GMRS %s' % x for x in range(1, 8)] + \
       ['GMRS %s' % x for x in range(15, 23)] + \
       ['GMRS Repeater %s' % x for x in range(15, 23)]
MAIN = ['Up', 'Down']
MODES = ['FM', 'NFM']
MONI = ['Squelch Off Momentarily', 'Squelch Off']
MRBANK = ['Bank %s' % x for x in range(1, 10)] + ['Bank 0']
MURS = ['MURS %s' % x for x in range(1, 6)]
NOAA = ['Weather Off', 'Weather On', 'Weather Alerts']
NOAACH = ['WX %s' % x for x in range(1, 8)]
PART95 = ['Normal(Part 90)', 'GMRS(Part 95A)', 'MURS(Part 95J)']
PAUSE = ['%s Seconds (TO)' % x for x in range(5, 20, 5)] + ['2 Seconds (CO)']
PFKEYT = ['Off', 'VOLT', 'CALL', 'FHSS', 'SUB PTT', 'ALARM', 'MONI']
PFKEYO = ['Off', 'VOLT', 'CALL', 'SUB PTT', 'ALARM', 'MONI']
POWER_LEVELS = [chirp_common.PowerLevel("High", watts=5),
                chirp_common.PowerLevel("Mid", watts=2),
                chirp_common.PowerLevel("Low", watts=1)]
SAVE = ['Off', '1:2', '1:3', '1:5', '1:8', 'Auto']
SQUELCH = ['%s' % x for x in range(0, 10)]
STOP = ['%s Seconds' % x for x in range(0, 4)] + ['Manual']
TAIL = ['Off', '120 Degree', '180 Degree', '240 Degree']
TBST = ['Off', '1750 Hz', '2100 Hz', '1000 Hz', '1450 Hz']
TMODES = ['', 'Tone', 'DTCS', '']
TONES = [62.5] + list(chirp_common.TONES)
TOT = ['Off'] + ['%s Seconds' % x for x in range(10, 280, 10)]
VDISPLAY = ['Frequency/Channel', 'Battery Voltage', 'Off']
VFOMR = ["VFO", "MR"]
VOXLEVEL = ['Off'] + ['%s' % x for x in range(1, 11)]
VOXDELAY = ['%.1f Seconds' % (0.1 * x) for x in range(5, 31)]
WORKMODE = ["Channel", "Bank"]


@directory.register
class AnyToneTERMN8RRadio(chirp_common.CloneModeRadio,
                          chirp_common.ExperimentalRadio):
    """AnyTone TERMN-8R"""
    VENDOR = "AnyTone"
    MODEL = "TERMN-8R"
    BAUD_RATE = 9600
    NEEDS_COMPAT_SERIAL = False
    _file_ident = b"TERMN8R"

    # May try to mirror the OEM behavior later
    _ranges = [
        (0x0000, 0x8000),
        ]

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = ("The Anytone driver is currently experimental. "
                           "There are no known issues with it, but you should "
                           "proceed with caution.")
        return rp

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_bank = False
        rf.has_cross = True
        rf.valid_tuning_steps = [2.5, 5, 6.25, 10, 12.5, 20, 25, 30, 50]
        rf.has_tuning_step = False
        rf.has_rx_dtcs = True
        rf.valid_skips = ["", "S"]
        rf.valid_modes = MODES
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'DTCS', 'Cross']
        rf.valid_cross_modes = ["Tone->Tone", "Tone->DTCS", "DTCS->Tone",
                                "->Tone", "->DTCS", "DTCS->", "DTCS->DTCS"]
        rf.valid_tones = TONES
        rf.valid_dtcs_codes = chirp_common.ALL_DTCS_CODES
        rf.valid_bands = [(136000000, 174000000),
                          (400000000, 520000000)]
        rf.valid_characters = CHARSET
        rf.valid_name_length = 6
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_duplexes = DUPLEXES
        rf.can_odd_split = True
        rf.memory_bounds = (0, 199)
        return rf

    def sync_in(self):
        self._mmap = _download(self)
        self.process_mmap()

    def sync_out(self):
        _upload(self)

    def process_mmap(self):
        self._memobj = bitwise.parse(mem_format, self._mmap)

    def _get_dcs_index(self, _mem, which):
        base = getattr(_mem, '%scode' % which)
        extra = getattr(_mem, '%sdcsextra' % which)
        return (int(extra) << 8) | int(base)

    def _set_dcs_index(self, _mem, which, index):
        base = getattr(_mem, '%scode' % which)
        extra = getattr(_mem, '%sdcsextra' % which)
        base.set_value(index & 0xFF)
        extra.set_value(index >> 8)

    def get_memory(self, number):
        bitpos = (1 << (number % 8))
        bytepos = (number / 8)

        _mem = self._memobj.memory[number]
        _skp = self._memobj.skip_flags[bytepos]
        _usd = self._memobj.used_flags[bytepos]

        mem = chirp_common.Memory()
        mem.number = number

        if _usd & bitpos:
            mem.empty = True
            return mem

        mem.freq = int(_mem.freq) * 100

        # compensate for 6.25 and 12.5 kHz tuning steps, add 500 Hz if needed
        lastdigit = int(_mem.freq) % 10
        if (lastdigit == 2 or lastdigit == 7):
            mem.freq += 50

        mem.offset = int(_mem.offset) * 100
        mem.name = self.filter_name(str(_mem.name).rstrip())
        mem.duplex = DUPLEXES[_mem.duplex]
        mem.mode = MODES[_mem.channel_width]

        if bool(_mem.tx_off):
            mem.duplex = "off"
            mem.offset = 0

        rxtone = txtone = None
        rxmode = TMODES[_mem.rxtmode]
        txmode = TMODES[_mem.txtmode]
        if txmode == "Tone":
            txtone = TONES[_mem.txtone]
        elif txmode == "DTCS":
            txtone = chirp_common.ALL_DTCS_CODES[self._get_dcs_index(_mem,
                                                                     'tx')]
        if rxmode == "Tone":
            rxtone = TONES[_mem.rxtone]
        elif rxmode == "DTCS":
            rxtone = chirp_common.ALL_DTCS_CODES[self._get_dcs_index(_mem,
                                                                     'rx')]

        rxpol = _mem.rxinv and "R" or "N"
        txpol = _mem.txinv and "R" or "N"

        chirp_common.split_tone_decode(mem,
                                       (txmode, txtone, txpol),
                                       (rxmode, rxtone, rxpol))

        if _skp & bitpos:
            mem.skip = "S"

        mem.power = POWER_LEVELS[_mem.power]

        mem.extra = RadioSettingGroup("Extra", "extra")

        rs = RadioSetting("bcl", "Busy Channel Lockout",
                          RadioSettingValueList(BCLO,
                                                BCLO[_mem.bcl]))
        mem.extra.append(rs)

        rs = RadioSetting("squelch", "Squelch",
                          RadioSettingValueList(SQUELCH,
                                                SQUELCH[_mem.squelch]))
        mem.extra.append(rs)

        return mem

    def set_memory(self, mem):
        bitpos = (1 << (mem.number % 8))
        bytepos = (mem.number / 8)

        _mem = self._memobj.memory[mem.number]
        _skp = self._memobj.skip_flags[bytepos]
        _usd = self._memobj.used_flags[bytepos]

        if mem.empty:
            _usd |= bitpos
            _skp |= bitpos
            _mem.set_raw("\xFF" * 32)
            return
        _usd &= ~bitpos

        if _mem.get_raw(asbytes=False) == ("\xFF" * 32):
            LOG.debug("Initializing empty memory")
            _mem.set_raw("\x00" * 32)
            _mem.squelch = 3

        _mem.freq = mem.freq / 100

        if mem.duplex == "off":
            _mem.duplex = DUPLEXES.index("")
            _mem.offset = 0
            _mem.tx_off = True
        elif mem.duplex == "split":
            diff = mem.offset - mem.freq
            _mem.duplex = DUPLEXES.index("-") if diff < 0 \
                else DUPLEXES.index("+")
            _mem.offset = abs(diff) / 100
        else:
            _mem.offset = mem.offset / 100
            _mem.duplex = DUPLEXES.index(mem.duplex)

        _mem.name = mem.name.ljust(6)

        try:
            _mem.channel_width = MODES.index(mem.mode)
        except ValueError:
            _mem.channel_width = 0

        ((txmode, txtone, txpol),
         (rxmode, rxtone, rxpol)) = chirp_common.split_tone_encode(mem)

        _mem.txtmode = TMODES.index(txmode)
        _mem.rxtmode = TMODES.index(rxmode)
        if txmode == "Tone":
            _mem.txtone = TONES.index(txtone)
        elif txmode == "DTCS":
            self._set_dcs_index(_mem, 'tx',
                                chirp_common.ALL_DTCS_CODES.index(txtone))
        if rxmode == "Tone":
            _mem.sqlmode = 1
            _mem.rxtone = TONES.index(rxtone)
        elif rxmode == "DTCS":
            _mem.sqlmode = 1
            self._set_dcs_index(_mem, 'rx',
                                chirp_common.ALL_DTCS_CODES.index(rxtone))
        else:
            _mem.sqlmode = 0

        _mem.txinv = txpol == "R"
        _mem.rxinv = rxpol == "R"

        if mem.skip:
            _skp |= bitpos
        else:
            _skp &= ~bitpos

        if mem.power:
            _mem.power = POWER_LEVELS.index(mem.power)
        else:
            _mem.power = 0

        for setting in mem.extra:
            setattr(_mem, setting.get_name(), setting.value)

    def get_settings(self):
        _msg = self._memobj.welcome_msg
        _oem = self._memobj.oem_info
        _settings = self._memobj.settings
        cfg_grp = RadioSettingGroup("cfg_grp", "Function Setup")
        oem_grp = RadioSettingGroup("oem_grp", "OEM Info")

        group = RadioSettings(cfg_grp,
                              oem_grp)

        def _filter(name):
            filtered = ""
            for char in str(name):
                if char in chirp_common.CHARSET_ASCII:
                    filtered += char
                else:
                    filtered += " "
            return filtered

        #
        # Function Setup
        #

        rs = RadioSetting("welcome_msg.line1", "Welcome Message 1",
                          RadioSettingValueString(
                              0, 6, _filter(_msg.line1)))
        cfg_grp.append(rs)

        rs = RadioSetting("welcome_msg.line2", "Welcome Message 2",
                          RadioSettingValueString(
                              0, 6, _filter(_msg.line2)))
        cfg_grp.append(rs)

        rs = RadioSetting("display", "Display Mode",
                          RadioSettingValueList(DISPLAY,
                                                DISPLAY[_settings.display]))
        cfg_grp.append(rs)

        rs = RadioSetting("upvfomr", "Up VFO/MR",
                          RadioSettingValueList(VFOMR,
                                                VFOMR[_settings.upvfomr]))
        cfg_grp.append(rs)

        rs = RadioSetting("dnvfomr", "Down VFO/MR",
                          RadioSettingValueList(VFOMR,
                                                VFOMR[_settings.dnvfomr]))
        cfg_grp.append(rs)

        rs = RadioSetting("upwork", "Up Work Mode",
                          RadioSettingValueList(WORKMODE,
                                                WORKMODE[_settings.upwork]))
        cfg_grp.append(rs)

        rs = RadioSetting("upmrbank", "Up MR Bank",
                          RadioSettingValueList(MRBANK,
                                                MRBANK[_settings.upmrbank]))
        cfg_grp.append(rs)

        rs = RadioSetting("upmrch", "Up MR Channel",
                          RadioSettingValueInteger(0, 200, _settings.upmrch))
        cfg_grp.append(rs)

        rs = RadioSetting("dnwork", "Down Work Mode",
                          RadioSettingValueList(WORKMODE,
                                                WORKMODE[_settings.dnwork]))
        cfg_grp.append(rs)

        rs = RadioSetting("dnmrbank", "Down MR Bank",
                          RadioSettingValueList(MRBANK,
                                                MRBANK[_settings.dnmrbank]))
        cfg_grp.append(rs)

        rs = RadioSetting("dnmrch", "Down MR Channel",
                          RadioSettingValueInteger(0, 200, _settings.dnmrch))
        cfg_grp.append(rs)

        rs = RadioSetting("main", "Main",
                          RadioSettingValueList(MAIN,
                                                MAIN[_settings.main]))
        cfg_grp.append(rs)

        rs = RadioSetting("pause", "Scan Pause Time",
                          RadioSettingValueList(PAUSE,
                                                PAUSE[_settings.pause]))
        cfg_grp.append(rs)

        rs = RadioSetting("stop", "Function Keys Stop Time",
                          RadioSettingValueList(STOP,
                                                STOP[_settings.stop]))
        cfg_grp.append(rs)

        rs = RadioSetting("backlight", "Backlight",
                          RadioSettingValueList(BACKLIGHT,
                                                BACKLIGHT[
                                                    _settings.backlight]))
        cfg_grp.append(rs)

        rs = RadioSetting("color", "Backlight Color",
                          RadioSettingValueList(COLOR,
                                                COLOR[_settings.color]))
        cfg_grp.append(rs)

        rs = RadioSetting("vdisplay", "Vice-Machine Display",
                          RadioSettingValueList(VDISPLAY,
                                                VDISPLAY[_settings.vdisplay]))
        cfg_grp.append(rs)

        rs = RadioSetting("voxlevel", "Vox Level",
                          RadioSettingValueList(VOXLEVEL,
                                                VOXLEVEL[_settings.voxlevel]))
        cfg_grp.append(rs)

        rs = RadioSetting("voxdelay", "Vox Delay",
                          RadioSettingValueList(VOXDELAY,
                                                VOXDELAY[_settings.voxdelay]))
        cfg_grp.append(rs)

        rs = RadioSetting("tot", "Time Out Timer",
                          RadioSettingValueList(TOT,
                                                TOT[_settings.tot]))
        cfg_grp.append(rs)

        rs = RadioSetting("tbst", "Tone Burst",
                          RadioSettingValueList(TBST,
                                                TBST[_settings.tbst]))
        cfg_grp.append(rs)

        rs = RadioSetting("monikey", "MONI Key Function",
                          RadioSettingValueList(MONI,
                                                MONI[_settings.monikey]))
        cfg_grp.append(rs)

        if self.MODEL == "TERMN-8R":
            rs = RadioSetting("pf1key", "PF1 Key Function",
                              RadioSettingValueList(PFKEYT,
                                                    PFKEYT[_settings.pf1key]))
            cfg_grp.append(rs)

            rs = RadioSetting("pf2key", "PF2 Key Function",
                              RadioSettingValueList(PFKEYT,
                                                    PFKEYT[_settings.pf2key]))
            cfg_grp.append(rs)

        if self.MODEL == "OBLTR-8R":
            rs = RadioSetting("pf1key", "PF1 Key Function",
                              RadioSettingValueList(PFKEYO,
                                                    PFKEYO[_settings.pf1key]))
            cfg_grp.append(rs)

            rs = RadioSetting("fmam", "PF2 Key Function",
                              RadioSettingValueList(PFKEYO,
                                                    PFKEYO[_settings.fmam]))
            cfg_grp.append(rs)

        rs = RadioSetting("apo", "Automatic Power Off",
                          RadioSettingValueList(APO,
                                                APO[_settings.apo]))
        cfg_grp.append(rs)

        rs = RadioSetting("save", "Power Save",
                          RadioSettingValueList(SAVE,
                                                SAVE[_settings.save]))
        cfg_grp.append(rs)

        rs = RadioSetting("tail", "Tail Eliminator Type",
                          RadioSettingValueList(TAIL,
                                                TAIL[_settings.tail]))
        cfg_grp.append(rs)

        rs = RadioSetting("fmvfomr", "FM VFO/MR",
                          RadioSettingValueList(VFOMR,
                                                VFOMR[_settings.fmvfomr]))
        cfg_grp.append(rs)

        rs = RadioSetting("fmmrch", "FM MR Channel",
                          RadioSettingValueInteger(0, 100, _settings.fmmrch))
        cfg_grp.append(rs)

        rs = RadioSetting("noaa", "NOAA",
                          RadioSettingValueList(NOAA,
                                                NOAA[_settings.noaa]))
        cfg_grp.append(rs)

        rs = RadioSetting("noaach", "NOAA Channel",
                          RadioSettingValueList(NOAACH,
                                                NOAACH[_settings.noaach]))
        cfg_grp.append(rs)

        rs = RadioSetting("part95", "PART95",
                          RadioSettingValueList(PART95,
                                                PART95[_settings.part95]))
        cfg_grp.append(rs)

        rs = RadioSetting("gmrs", "GMRS",
                          RadioSettingValueList(GMRS,
                                                GMRS[_settings.gmrs]))
        cfg_grp.append(rs)

        rs = RadioSetting("murs", "MURS",
                          RadioSettingValueList(MURS,
                                                MURS[_settings.murs]))
        cfg_grp.append(rs)

        for i in range(0, 9):
            val = self._memobj.banklink[i].bank
            rs = RadioSetting("banklink/%i.bank" % i,
                              "Bank Link %i" % (i + 1),
                              RadioSettingValueBoolean(val))
            cfg_grp.append(rs)

        val = self._memobj.banklink[9].bank
        rs = RadioSetting("banklink/9.bank", "Bank Link 0",
                          RadioSettingValueBoolean(val))
        cfg_grp.append(rs)

        rs = RadioSetting("allband", "All Band",
                          RadioSettingValueBoolean(_settings.allband))
        cfg_grp.append(rs)

        rs = RadioSetting("alarmoff", "Alarm Function Off",
                          RadioSettingValueBoolean(_settings.alarmoff))
        cfg_grp.append(rs)

        rs = RadioSetting("beep", "Beep",
                          RadioSettingValueBoolean(_settings.beep))
        cfg_grp.append(rs)

        rs = RadioSetting("radio", "Radio",
                          RadioSettingValueBoolean(_settings.radio))
        cfg_grp.append(rs)

        if self.MODEL == "TERMN-8R":
            rs = RadioSetting("keylock", "Keylock",
                              RadioSettingValueBoolean(_settings.keylock))
            cfg_grp.append(rs)

            rs = RadioSetting("fastscan", "Fast Scan",
                              RadioSettingValueBoolean(_settings.fastscan))
            cfg_grp.append(rs)

        if self.MODEL == "OBLTR-8R":
            # "pf2key" is used for OBLTR-8R "keylock"
            rs = RadioSetting("pf2key", "Keylock",
                              RadioSettingValueBoolean(_settings.pf2key))
            cfg_grp.append(rs)

            rs = RadioSetting("fastscano", "Fast Scan",
                              RadioSettingValueBoolean(_settings.fastscano))
            cfg_grp.append(rs)

        rs = RadioSetting("uplink", "Up Bank Link Select",
                          RadioSettingValueBoolean(_settings.uplink))
        cfg_grp.append(rs)

        rs = RadioSetting("downlink", "Down Bank Link Select",
                          RadioSettingValueBoolean(_settings.downlink))
        cfg_grp.append(rs)

        #
        # OEM info
        #

        val = RadioSettingValueString(0, 7, _filter(_oem.model))
        val.set_mutable(False)
        rs = RadioSetting("oem_info.model", "Model", val)
        oem_grp.append(rs)

        val = RadioSettingValueString(0, 9, _filter(_oem.date))
        val.set_mutable(False)
        rs = RadioSetting("oem_info.date", "Date", val)
        oem_grp.append(rs)

        val = RadioSettingValueString(0, 16, _filter(_oem.dealer))
        val.set_mutable(False)
        rs = RadioSetting("oem_info.dealer", "Dealer Code", val)
        oem_grp.append(rs)

        val = RadioSettingValueString(0, 9, _filter(_oem.stockdate))
        val.set_mutable(False)
        rs = RadioSetting("oem_info.stockdate", "Stock Date", val)
        oem_grp.append(rs)

        val = RadioSettingValueString(0, 9, _filter(_oem.selldate))
        val.set_mutable(False)
        rs = RadioSetting("oem_info.selldate", "Sell Date", val)
        oem_grp.append(rs)

        return group

    def set_settings(self, settings):
        _settings = self._memobj.settings
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    name = element.get_name()
                    if "." in name:
                        bits = name.split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            if "/" in bit:
                                bit, index = bit.split("/", 1)
                                index = int(index)
                                obj = getattr(obj, bit)[index]
                            else:
                                obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = _settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("Using apply callback")
                        element.run_apply_callback()
                    else:
                        LOG.debug("Setting %s = %s" % (setting, element.value))
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise

    @classmethod
    def match_model(cls, filedata, filename):
        return cls._file_ident in filedata[0x10:0x20]


@directory.register
class AnyToneOBLTR8RRadio(AnyToneTERMN8RRadio):
    """AnyTone OBLTR-8R"""
    VENDOR = "AnyTone"
    MODEL = "OBLTR-8R"
    _file_ident = b"OBLTR8R"



# Copyright 2017 SASANO Takayoshi (JG1UAA) <uaa@uaa.org.uk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp.drivers import icf
from chirp import chirp_common, directory, bitwise

# memory number:
# 000 -  999	regular memory channels (supported, others not)
# 1000 - 1049	scan edges
# 1050 - 1249	auto write channels
# 1250		call channel (C0)
# 1251		call channel (C1)


MEM_FORMAT = """
struct {
  ul32 freq;
  ul32 offset;
  ul16 train_sql:2,
       tmode:3,
       duplex:2,
       train_tone:9;
  ul16 tuning_step:4,
       rtone:6,
       ctone:6;
  ul16 unknown0:6,
       mode:3,
       dtcs:7;
  u8   unknown1:6,
       dtcs_polarity:2;
  char name[6];
} memory[1251];

#seekto 0x6b1e;
struct {
  u8 bank;
  u8 index;
} banks[1050];

#seekto 0x689e;
u8 used[132];

#seekto 0x6922;
u8 skips[132];

#seekto 0x69a6;
u8 pskips[132];

#seekto 0x7352;
struct {
  char name[6];
} bank_names[18];

"""

MODES = ["FM", "WFM", "AM", "Auto"]
TMODES = ["", "Tone", "TSQL", "", "DTCS"]
DUPLEX = ["", "-", "+"]
DTCS_POLARITY = ["NN", "NR", "RN", "RR"]
TUNING_STEPS = [5.0, 6.25, 8.33, 9.0, 10.0, 12.5, 15.0, 20.0,
                25.0, 30.0, 50.0, 100.0, 200.0, 0.0]  # 0.0 as "Auto"


class ICP7Bank(icf.IcomBank):
    """ICP7 bank"""
    def get_name(self):
        _bank = self._model._radio._memobj.bank_names[self.index]
        return str(_bank.name).rstrip()

    def set_name(self, name):
        _bank = self._model._radio._memobj.bank_names[self.index]
        _bank.name = name.ljust(6)[:6]


@directory.register
class ICP7Radio(icf.IcomCloneModeRadio):
    """Icom IC-P7"""
    VENDOR = "Icom"
    MODEL = "IC-P7"

    _model = "\x28\x69\x00\x01"
    _memsize = 0x7500
    _endframe = "Icom Inc\x2e\x41\x38"

    _ranges = [(0x0000, 0x7500, 32)]

    _num_banks = 18
    _bank_class = ICP7Bank
    _can_hispeed = True

    def _get_bank(self, loc):
        _bank = self._memobj.banks[loc]
        if _bank.bank != 0xff:
            return _bank.bank
        else:
            return None

    def _set_bank(self, loc, bank):
        _bank = self._memobj.banks[loc]
        if bank is None:
            _bank.bank = 0xff
        else:
            _bank.bank = bank

    def _get_bank_index(self, loc):
        _bank = self._memobj.banks[loc]
        return _bank.index

    def _set_bank_index(self, loc, index):
        _bank = self._memobj.banks[loc]
        _bank.index = index

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.memory_bounds = (0, 999)
        rf.valid_tmodes = TMODES
        rf.valid_duplexes = DUPLEX
        rf.valid_modes = MODES
        rf.valid_bands = [(495000, 999990000)]
        rf.valid_skips = ["", "S", "P"]
        rf.valid_tuning_steps = TUNING_STEPS
        rf.valid_name_length = 6
        rf.has_settings = True
        rf.has_ctone = True
        rf.has_bank = True
        rf.has_bank_index = True
        rf.has_bank_names = True
        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number])

    def get_memory(self, number):
        bit = 1 << (number % 8)
        byte = int(number / 8)

        _mem = self._memobj.memory[number]
        _usd = self._memobj.used[byte]
        _skp = self._memobj.skips[byte]
        _psk = self._memobj.pskips[byte]

        mem = chirp_common.Memory()
        mem.number = number

        if _usd & bit:
            mem.empty = True
            return mem

        mem.freq = _mem.freq // 3
        mem.offset = _mem.offset // 3
        mem.tmode = TMODES[_mem.tmode]
        mem.duplex = DUPLEX[_mem.duplex]
        mem.tuning_step = TUNING_STEPS[_mem.tuning_step]
        mem.rtone = chirp_common.TONES[_mem.rtone]
        mem.ctone = chirp_common.TONES[_mem.ctone]
        mem.mode = MODES[_mem.mode]
        mem.dtcs = chirp_common.DTCS_CODES[_mem.dtcs]
        mem.dtcs_polarity = DTCS_POLARITY[_mem.dtcs_polarity]
        mem.name = str(_mem.name).rstrip()

        if _skp & bit:
            mem.skip = "P" if _psk & bit else "S"
        else:
            mem.skip = ""

        return mem

    def set_memory(self, mem):
        bit = 1 << (mem.number % 8)
        byte = int(mem.number / 8)

        _mem = self._memobj.memory[mem.number]
        _usd = self._memobj.used[byte]
        _skp = self._memobj.skips[byte]
        _psk = self._memobj.pskips[byte]

        if mem.empty:
            _usd |= bit

            # We use default value instead of zero-fill
            # to avoid unexpected behavior.
            _mem.freq = 15000
            _mem.offset = 479985000
            _mem.train_sql = ~0
            _mem.tmode = ~0
            _mem.duplex = ~0
            _mem.train_tone = ~0
            _mem.tuning_step = ~0
            _mem.rtone = ~0
            _mem.ctone = ~0
            _mem.unknown0 = 0
            _mem.mode = ~0
            _mem.dtcs = ~0
            _mem.unknown1 = ~0
            _mem.dtcs_polarity = ~0
            _mem.name = "      "

            _skp |= bit
            _psk |= bit

        else:
            _usd &= ~bit

            _mem.freq = mem.freq * 3
            _mem.offset = mem.offset * 3
            _mem.train_sql = 0  # Train SQL mode (0:off 1:Tone 2:MSK)
            _mem.tmode = TMODES.index(mem.tmode)
            _mem.duplex = DUPLEX.index(mem.duplex)
            _mem.train_tone = 228  # Train SQL Tone (x10 Hz)
            _mem.tuning_step = TUNING_STEPS.index(mem.tuning_step)
            _mem.rtone = chirp_common.TONES.index(mem.rtone)
            _mem.ctone = chirp_common.TONES.index(mem.ctone)
            _mem.unknown0 = 0  # unknown (always zero)
            _mem.mode = MODES.index(mem.mode)
            _mem.dtcs = chirp_common.DTCS_CODES.index(mem.dtcs)
            _mem.unknown1 = ~0  # unknown (always one)
            _mem.dtcs_polarity = DTCS_POLARITY.index(mem.dtcs_polarity)
            _mem.name = mem.name.ljust(6)[:6]

            if mem.skip == "S":
                _skp |= bit
                _psk &= ~bit
            elif mem.skip == "P":
                _skp |= bit
                _psk |= bit
            else:
                _skp &= ~bit
                _psk &= ~bit



# Copyright 2022 Jim Unroe <rock.unroe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
import logging

from chirp import chirp_common, directory, memmap
from chirp import bitwise, errors, util
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueInteger, RadioSettingValueList, \
    RadioSettingValueBoolean, RadioSettingValueFloat, \
    RadioSettings

LOG = logging.getLogger(__name__)

MEM_FORMAT = """
#seekto 0x0010;
struct {
  u8 range174_180;  // 174-180 MHz
  u8 range180_190;  // 180-190 MHz
  u8 range190_200;  // 190-200 MHz
  u8 range200_210;  // 200-210 MHz
  u8 range210_220;  // 210-220 MHz
  u8 range220_230;  // 220-230 MHz
  u8 range230_240;  // 230-240 MHz
  u8 range240_250;  // 240-250 MHz
  u8 range250_260;  // 240-250 MHz
  u8 range260_270;  // 260-270 MHz
  u8 range270_280;  // 270-280 MHz
  u8 range280_290;  // 280-290 MHz
  u8 range290_300;  // 290-300 MHz
  u8 range300_310;  // 300-310 MHz
  u8 range310_320;  // 310-320 MHz
  u8 range320_330;  // 320-330 MHz
  u8 range330_340;  // 330-340 MHz
  u8 range340_350;  // 340-350 MHz
  u8 range350_360;  // 350-360 MHz
  u8 range360_370;  // 360-370 MHz
  u8 range370_380;  // 370-380 MHz
  u8 range380_390;  // 380-390 MHz
  u8 range390_400;  // 290-400 MHz
  u8 range520_530;  // 520-530 MHz
  u8 range530_540;  // 530-540 MHz
  u8 range540_550;  // 540-550 MHz
  u8 range550_560;  // 550-560 MHz
  u8 range560_570;  // 560-570 MHz
  u8 range570_580;  // 570-580 MHz
  u8 range580_590;  // 580-590 MHz
  u8 range590_600;  // 590-600 MHz
  u8 range600_610;  // 600-610 MHz
  u8 range610_620;  // 610-620 MHz
  u8 range620_630;  // 620-630 MHz
  u8 range630_640;  // 630-640 MHz
  u8 range640_650;  // 640-650 MHz
  u8 range650_660;  // 650-660 MHz
} txallow;

#seekto 0x0039;
struct {
  u8 display_timer;         // 0x0039        Display Timer
  u8 led_timer;             // 0x003A        LED Timer
  u8 auto_lock;             // 0x003B        Auto Lock
  u8 pfkey1short;           // 0x003C        FS1 Short
  u8 pfkey1long;            // 0x003D        FS1 Long
  u8 pfkey2short;           // 0x003E        FS2 Short
  u8 pfkey2long;            // 0x003F        FS2 Long
  u8 unknown_0;             // 0x0040
  u8 voice:2,               // 0x0041        Voice Prompt
     beep:1,                //               Beep Switch
     roger:2,               //               Roger Tone
     save:3;                //               Power Save
  u8 dispmode:1,            // 0x0042        Display Mode
     dstandby:1,            //               Dual Standby
     unknown_1:1,
     standby:1,             //               Radio Standby
     squelch:4;             //               Squelch Level
  u8 vox_level:4,           // 0x0043        VOX Level
     vox_delay:4;           //               VOX Delay
  u8 brightness:2,          // 0x0044        Screen Light
     unknown_2:1,
     alarm:1,               //               Alarm
     scramble:4;            //               Scramble
  u8 screen:1,              // 0x0045        Screen On-off
     led:1,                 //               LED On-off
     key_lock:1,            //               Key Lock
     radio:1,               //               Radio On-off
     compander:1,           //               Compander
     rp:1,                  //               Offset
     freq_lock:1,           //               Freq Lock
     vox:1;                 //               VOX On-off
  u8 unknown_4:4,           // 0x0046
     txpri:1,               //               TX PRI
     unknown_5:1,
     tail_tone:2;           //               Tail Tone
  ul16 fmcur;               // 0x0047        Radio Freq
  u8 unknown_49;            // 0x0049
  u8 unknown_4A;            // 0x004A
  u8 unknown_4B;            // 0x004B
  u8 unknown_4C;            // 0x004C
  u8 dsrangea;              // 0x004D        Dual Standby Range A
  u8 dsrangeb;              // 0x004E        Dual Standby Range B
  u8 tot;                   // 0x004F
} settings;

struct memory {
  ul32 rxfreq;                               // 00-03
  ul16 rx_tone;   // PL/DPL Decode           // 04-05
  ul32 txfreq;                               // 06-09
  ul16 tx_tone;   // PL/DPL Encode           // 0a-0b
  u8 lowpower:1,  // Power Level             // 0c
     isnarrow:1,  // Bandwidth
     unknown1:1,
     unknown2:3,
     bcl:1,       // Busy Channel Lockout
     scan:1;      // Scan Add
  u8 unknown3[3];                            // 0d-0f
};

// #seekto 0x0050;
struct memory channels[128];

"""

CMD_ACK = b"\x06"

RB15_DTCS = tuple(sorted(chirp_common.DTCS_CODES + (645,)))

_STEP_LIST = [0.25, 1.25, 2.5, 5., 6.25, 10., 12.5, 25., 50., 100.]

LIST_ALARM = ["Local", "Remote"]
LIST_BRIGHT = ["Low", "Middle", "High"]
LIST_DISPM = ["Frequency", "Channel Number"]
LIST_ROGER = ["Off", "Start", "End", "Start and End"]
LIST_SAVE = ["Off", "1:1", "1:2", "1:3", "1:4", "1:5"]
LIST_SCRAMBLE = ["Off"] + ["%s" % x for x in range(1, 9)]
LIST_SKEY = ["None", "Monitor", "Local Alarm", "Remote Alarm",
             "Screen On/Off", "Frequency Lock"]
LIST_TIMER = ["Off"] + ["%s seconds" % x for x in range(15, 615, 15)]
LIST_TXPRI = ["Edit", "Busy"]
LIST_VOICE = ["Off", "Chinese", "English"]

TXALLOW_CHOICES = ["RX Only", "TX/RX"]
TXALLOW_VALUES = [0xFF, 0x00]


def _checksum(data):
    cs = 0
    for byte in data:
        cs += byte
    return cs % 256


def _enter_programming_mode(radio):
    serial = radio.pipe

    # lengthen the timeout here as these radios are resetting due to timeout
    radio.pipe.timeout = 0.75

    exito = False
    for i in range(0, 5):
        serial.write(radio.magic)
        ack = serial.read(1)

        try:
            if ack == CMD_ACK:
                exito = True
                break
        except:
            LOG.debug("Attempt #%s, failed, trying again" % i)
            pass

    # return timeout to default value
    radio.pipe.timeout = 0.25

    # check if we had EXITO
    if exito is False:
        msg = "The radio did not accept program mode after five tries.\n"
        msg += "Check you interface cable and power cycle your radio."
        raise errors.RadioError(msg)


def _exit_programming_mode(radio):
    serial = radio.pipe
    try:
        serial.write(b"93" + b"\x05\xEE\x5F")
    except:
        raise errors.RadioError("Radio refused to exit programming mode")


def _read_block(radio, block_addr, block_size):
    serial = radio.pipe

    cmd = struct.pack(">BH", ord(b'R'), block_addr + 0x0340)

    ccs = bytes([_checksum(cmd)])

    expectedresponse = b"R" + cmd[1:]

    cmd = cmd + ccs

    LOG.debug("Reading block %04x..." % (block_addr + 0x0340))

    try:
        serial.write(cmd)
        response = serial.read(3 + block_size + 1)

        cs = _checksum(response[:-1])

        if response[:3] != expectedresponse:
            raise Exception("Error reading block %04x." % (block_addr +
                            0x0340))

        chunk = response[3:]

        if chunk[-1] != cs:
            raise Exception("Block failed checksum!")

        block_data = chunk[:-1]
    except:
        raise errors.RadioError("Failed to read block at %04x" % block_addr +
                                0x0340)

    return block_data


def _write_block(radio, block_addr, block_size):
    serial = radio.pipe

    data = radio.get_mmap()[block_addr:block_addr + block_size]

    cmd = struct.pack(">BH", ord(b'W'), block_addr + 0x0340)

    cs = bytes([_checksum(cmd + data)])
    data += cs

    LOG.debug("Writing Data:")
    LOG.debug(util.hexprint(cmd + data))

    try:
        serial.write(cmd + data)
        if serial.read(1) != CMD_ACK:
            raise Exception("No ACK")
    except:
        raise errors.RadioError("Failed to send block "
                                "to radio at %04x" % (block_addr + 0x0340))


def do_download(radio):
    LOG.debug("download")
    _enter_programming_mode(radio)

    data = b""

    status = chirp_common.Status()
    status.msg = "Cloning from radio"

    status.cur = 0
    status.max = radio._memsize

    for addr in range(0, radio._memsize, radio.BLOCK_SIZE):
        status.cur = addr + radio.BLOCK_SIZE
        radio.status_fn(status)

        block = _read_block(radio, addr, radio.BLOCK_SIZE)
        data += block

        LOG.debug("Address: %04x" % addr)
        LOG.debug(util.hexprint(block))

    _exit_programming_mode(radio)

    return memmap.MemoryMapBytes(data)


def do_upload(radio):
    status = chirp_common.Status()
    status.msg = "Uploading to radio"

    _enter_programming_mode(radio)

    status.cur = 0
    status.max = radio._memsize

    for start_addr, end_addr in radio._ranges:
        for addr in range(start_addr, end_addr, radio.BLOCK_SIZE):
            status.cur = addr + radio.BLOCK_SIZE
            radio.status_fn(status)
            _write_block(radio, addr, radio.BLOCK_SIZE)

    _exit_programming_mode(radio)


def _split(rf, f1, f2):
    """Returns False if the two freqs are in the same band (no split)
    or True otherwise"""

    # determine if the two freqs are in the same band
    for low, high in rf.valid_bands:
        if f1 >= low and f1 <= high and \
                f2 >= low and f2 <= high:
            # if the two freqs are on the same Band this is not a split
            return False

    # if you get here is because the freq pairs are split
    return True


class IradioUV5118(chirp_common.CloneModeRadio):
    """IRADIO UV5118"""
    VENDOR = "Iradio"
    MODEL = "UV-5118"
    BAUD_RATE = 9600
    NEEDS_COMPAT_SERIAL = False

    BLOCK_SIZE = 0x10
    magic = b"93" + b"\x05\x10\x81"

    VALID_BANDS = [(20000000, 64000000),  # RX only
                   (108000000, 136000000),  # RX only (Air Band)
                   (136000000, 174000000),  # TX/RX (VHF)
                   (174000000, 400000000),  # TX/RX
                   (400000000, 520000000),  # TX/RX (UHF)
                   (520000000, 660000000)]  # TX/RX

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=2.00),
                    chirp_common.PowerLevel("Low", watts=0.50)]

    # Radio's memory starts at 0x0340
    # Radio's memory ends at 0x0B90

    _ranges = [
               (0x0000, 0x0850),
              ]
    _memsize = 0x0850

    _upper = 128

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_bank = False
        rf.has_ctone = True
        rf.has_cross = True
        rf.has_rx_dtcs = True
        rf.has_tuning_step = False
        rf.can_odd_split = True
        rf.has_name = False
        rf.valid_skips = ["", "S"]
        rf.valid_tmodes = ["", "Tone", "TSQL", "DTCS", "Cross"]
        rf.valid_cross_modes = ["Tone->Tone", "Tone->DTCS", "DTCS->Tone",
                                "->Tone", "->DTCS", "DTCS->", "DTCS->DTCS"]
        rf.valid_power_levels = self.POWER_LEVELS
        rf.valid_duplexes = ["", "-", "+", "split"]
        rf.valid_modes = ["FM", "NFM"]  # 25 kHz, 12.5 kHz.
        rf.valid_dtcs_codes = RB15_DTCS
        rf.memory_bounds = (1, self._upper)
        rf.valid_tuning_steps = _STEP_LIST
        rf.valid_bands = self.VALID_BANDS

        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def sync_in(self):
        """Download from radio"""
        try:
            data = do_download(self)
        except errors.RadioError:
            # Pass through any real errors we raise
            raise
        except:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during download')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')
        self._mmap = data
        self.process_mmap()

    def sync_out(self):
        """Upload to radio"""
        try:
            do_upload(self)
        except:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during upload')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number - 1])

    @staticmethod
    def _decode_tone(toneval):
        # DCS examples:
        # D023N - 1013 - 0001 0000 0001 0011
        #                   ^-DCS
        # D023I - 2013 - 0010 0000 0001 0100
        #                  ^--DCS inverted
        # D754I - 21EC - 0010 0001 1110 1100
        #    code in octal-------^^^^^^^^^^^

        if toneval == 0x3000:
            return '', None, None
        elif toneval & 0x1000:
            # DTCS N
            code = int('%o' % (toneval & 0x1FF))
            return 'DTCS', code, 'N'
        elif toneval & 0x2000:
            # DTCS R
            code = int('%o' % (toneval & 0x1FF))
            return 'DTCS', code, 'R'
        else:
            return 'Tone', toneval / 10.0, None

    @staticmethod
    def _encode_tone(mode, val, pol):
        if not mode:
            return 0x3000
        elif mode == 'Tone':
            return int(val * 10)
        elif mode == 'DTCS':
            code = int('%i' % val, 8)
            if pol == 'N':
                code |= 0x1800
            if pol == 'R':
                code |= 0x2800
            return code
        else:
            raise errors.RadioError('Unsupported tone mode %r' % mode)

    def get_memory(self, number):
        mem = chirp_common.Memory()
        _mem = self._memobj.channels[number - 1]
        mem.number = number

        mem.freq = int(_mem.rxfreq) * 10

        # We'll consider any blank (i.e. 0 MHz frequency) to be empty
        if mem.freq == 0:
            mem.empty = True
            return mem

        if _mem.rxfreq.get_raw() == b"\xFF\xFF\xFF\xFF":
            mem.freq = 0
            mem.empty = True
            return mem

        if _mem.get_raw() == (b"\xFF" * 16):
            LOG.debug("Initializing empty memory")
            _mem.set_raw("\xFF" * 4 + "\x00\x30" + "\xFF" * 4 + "\x00\x30" +
                         "\x00" * 4)

        # Freq and offset
        mem.freq = int(_mem.rxfreq) * 10
        # TX freq set
        offset = (int(_mem.txfreq) * 10) - mem.freq
        if offset != 0:
            if _split(self.get_features(), mem.freq, int(
                      _mem.txfreq) * 10):
                mem.duplex = "split"
                mem.offset = int(_mem.txfreq) * 10
            elif offset < 0:
                mem.offset = abs(offset)
                mem.duplex = "-"
            elif offset > 0:
                mem.offset = offset
                mem.duplex = "+"
        else:
            mem.offset = 0

        mem.mode = _mem.isnarrow and "NFM" or "FM"

        chirp_common.split_tone_decode(mem,
                                       self._decode_tone(_mem.tx_tone),
                                       self._decode_tone(_mem.rx_tone))

        mem.power = self.POWER_LEVELS[_mem.lowpower]

        if not _mem.scan:
            mem.skip = "S"

        mem.extra = RadioSettingGroup("Extra", "extra")

        rs = RadioSetting("bcl", "BCL",
                          RadioSettingValueBoolean(_mem.bcl))
        mem.extra.append(rs)

        return mem

    def set_memory(self, mem):
        LOG.debug("Setting %i(%s)" % (mem.number, mem.extd_number))
        _mem = self._memobj.channels[mem.number - 1]

        # if empty memory
        if mem.empty:
            _mem.set_raw("\xFF" * 16)
            return

        _mem.set_raw("\xFF" * 4 + "\x00\x30" + "\xFF" * 4 + "\x00\x30" +
                     "\x00" * 4)

        _mem.rxfreq = mem.freq / 10

        if mem.duplex == "split":
            _mem.txfreq = mem.offset / 10
        elif mem.duplex == "+":
            _mem.txfreq = (mem.freq + mem.offset) / 10
        elif mem.duplex == "-":
            _mem.txfreq = (mem.freq - mem.offset) / 10
        else:
            _mem.txfreq = mem.freq / 10

        _mem.scan = mem.skip != "S"
        _mem.isnarrow = mem.mode == "NFM"

        txtone, rxtone = chirp_common.split_tone_encode(mem)
        _mem.tx_tone = self._encode_tone(*txtone)
        _mem.rx_tone = self._encode_tone(*rxtone)

        _mem.lowpower = mem.power == self.POWER_LEVELS[1]

        for setting in mem.extra:
            setattr(_mem, setting.get_name(), setting.value)

    def get_settings(self):
        _settings = self._memobj.settings
        _txallow = self._memobj.txallow
        basic = RadioSettingGroup("basic", "Basic Settings")
        sidekey = RadioSettingGroup("sidekey", "Side Key Settings")
        dwatch = RadioSettingGroup("dwatch", "Dual Watch Settings")
        txallow = RadioSettingGroup("txallow", "TX Allow")
        top = RadioSettings(basic, sidekey, dwatch, txallow)

        voice = RadioSetting("voice", "Language", RadioSettingValueList(
                             LIST_VOICE, LIST_VOICE[_settings.voice]))
        basic.append(voice)

        beep = RadioSetting("beep", "Key Beep",
                            RadioSettingValueBoolean(_settings.beep))
        basic.append(beep)

        roger = RadioSetting("roger", "Roger Tone",
                             RadioSettingValueList(
                                 LIST_ROGER, LIST_ROGER[_settings.roger]))
        basic.append(roger)

        save = RadioSetting("save", "Battery Save",
                            RadioSettingValueList(
                                LIST_SAVE, LIST_SAVE[_settings.save]))
        basic.append(save)

        dispmode = RadioSetting("dispmode", "Display Mode",
                                RadioSettingValueList(
                                    LIST_DISPM,
                                    LIST_DISPM[_settings.dispmode]))
        basic.append(dispmode)

        brightness = RadioSetting("brightness", "Screen Light",
                                  RadioSettingValueList(
                                      LIST_BRIGHT,
                                      LIST_BRIGHT[_settings.brightness]))
        basic.append(brightness)

        screen = RadioSetting("screen", "Screen",
                              RadioSettingValueBoolean(_settings.screen))
        basic.append(screen)

        display_timer = RadioSetting("display_timer", "Display Timer",
                                     RadioSettingValueList(
                                         LIST_TIMER,
                                         LIST_TIMER[_settings.display_timer]))
        basic.append(display_timer)

        led = RadioSetting("led", "LED",
                           RadioSettingValueBoolean(_settings.led))
        basic.append(led)

        led_timer = RadioSetting("led_timer", "LED Timer",
                                 RadioSettingValueList(
                                     LIST_TIMER,
                                     LIST_TIMER[_settings.led_timer]))
        basic.append(led_timer)

        squelch = RadioSetting("squelch", "Squelch Level",
                               RadioSettingValueInteger(
                                   0, 9, _settings.squelch))
        basic.append(squelch)

        vox = RadioSetting("vox", "VOX",
                           RadioSettingValueBoolean(_settings.vox))
        basic.append(vox)

        vox_level = RadioSetting("vox_level", "VOX Level",
                                 RadioSettingValueInteger(
                                     0, 9, _settings.vox_level))
        basic.append(vox_level)

        vox_delay = RadioSetting("vox_delay", "VOX Delay",
                                 RadioSettingValueInteger(
                                     0, 9, _settings.vox_delay))
        basic.append(vox_delay)

        compander = RadioSetting("compander", "Compander",
                                 RadioSettingValueBoolean(_settings.compander))
        basic.append(compander)

        scramble = RadioSetting("scramble", "Scramble",
                                RadioSettingValueList(
                                    LIST_SCRAMBLE,
                                    LIST_SCRAMBLE[_settings.scramble]))
        basic.append(scramble)

        rp = RadioSetting("rp", "Offset",
                          RadioSettingValueBoolean(_settings.rp))
        basic.append(rp)

        txpri = RadioSetting("txpri", "TX Priority",
                             RadioSettingValueList(
                                 LIST_TXPRI,
                                 LIST_TXPRI[_settings.txpri]))
        basic.append(txpri)

        tot = RadioSetting("tot", "Time-out Timer",
                           RadioSettingValueList(
                               LIST_TIMER,
                               LIST_TIMER[_settings.tot]))
        basic.append(tot)

        tail_tone = RadioSetting("tail_tone", "Tail Tone",
                                 RadioSettingValueInteger(
                                     0, 3, _settings.tail_tone))
        basic.append(tail_tone)

        alarm = RadioSetting("alarm", "Alarm",
                             RadioSettingValueList(
                                 LIST_ALARM,
                                 LIST_ALARM[_settings.alarm]))
        basic.append(alarm)

        freq_lock = RadioSetting("freq_lock", "Frequency Lock",
                                 RadioSettingValueBoolean(_settings.freq_lock))
        basic.append(freq_lock)

        key_lock = RadioSetting("key_lock", "Key Lock",
                                RadioSettingValueBoolean(_settings.key_lock))
        basic.append(key_lock)

        auto_lock = RadioSetting("auto_lock", "Auto Lock Timer",
                                 RadioSettingValueList(
                                     LIST_TIMER,
                                     LIST_TIMER[_settings.auto_lock]))
        basic.append(auto_lock)

        radio = RadioSetting("radio", "Broadcast FM Radio",
                             RadioSettingValueBoolean(_settings.radio))
        basic.append(radio)

        standby = RadioSetting("standby", "Radio Standby",
                               RadioSettingValueBoolean(_settings.standby))
        basic.append(standby)

        def myset_freq(setting, obj, atrb, mult):
            """ Callback to set frequency by applying multiplier"""
            value = int(float(str(setting.value)) * mult)
            setattr(obj, atrb, value)
            return

        # FM Broadcast Settings
        val = _settings.fmcur
        val = val / 10.0
        if val < 87.5 or val > 108.0:
            val = 90.4
        rx = RadioSettingValueFloat(87.5, 108.0, val, 0.1, 1)
        rset = RadioSetting("fmcur", "Broadcast FM Freq (MHz)", rx)
        rset.set_apply_callback(myset_freq, _settings, "fmcur", 10)
        basic.append(rset)

        # Side Key Settings
        pfkey1short = RadioSetting("pfkey1short", "Side Key 1 - Short Press",
                                   RadioSettingValueList(
                                       LIST_SKEY,
                                       LIST_SKEY[_settings.pfkey1short]))
        sidekey.append(pfkey1short)

        pfkey1long = RadioSetting("pfkey1long", "Side Key 1 - Long Press",
                                  RadioSettingValueList(
                                      LIST_SKEY,
                                      LIST_SKEY[_settings.pfkey1long]))
        sidekey.append(pfkey1long)

        pfkey2short = RadioSetting("pfkey2short", "Side Key 2 - Short Press",
                                   RadioSettingValueList(
                                       LIST_SKEY,
                                       LIST_SKEY[_settings.pfkey2short]))
        sidekey.append(pfkey2short)

        pfkey2long = RadioSetting("pfkey2long", "Side Key 2 - Long Press",
                                  RadioSettingValueList(
                                      LIST_SKEY,
                                      LIST_SKEY[_settings.pfkey2long]))
        sidekey.append(pfkey2long)

        dstandby = RadioSetting("dstandby", "Dual Standby",
                                RadioSettingValueBoolean(_settings.dstandby))
        dwatch.append(dstandby)

        dsrangea = RadioSetting("dsrangea", "Dual Standby Range A",
                                RadioSettingValueInteger(
                                    1, 128, _settings.dsrangea + 1))
        dwatch.append(dsrangea)

        dsrangeb = RadioSetting("dsrangeb", "Dual Standby Range B",
                                RadioSettingValueInteger(
                                    1, 128, _settings.dsrangeb + 1))
        dwatch.append(dsrangeb)

        def apply_txallow_listvalue(setting, obj):
            LOG.debug("Setting value: " + str(
                      setting.value) + " from list")
            val = str(setting.value)
            index = TXALLOW_CHOICES.index(val)
            val = TXALLOW_VALUES[index]
            obj.set_value(val)

        if _txallow.range174_180 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range174_180)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range174_180", "174-180 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range174_180)
        txallow.append(rset)

        if _txallow.range180_190 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range180_190)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range180_190", "180-190 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range180_190)
        txallow.append(rset)

        if _txallow.range190_200 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range190_200)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range190_200", "190-200 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range190_200)
        txallow.append(rset)

        if _txallow.range200_210 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range200_210)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range200_210", "200-210 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range200_210)
        txallow.append(rset)

        if _txallow.range210_220 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range210_220)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range210_220", "210-220 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range210_220)
        txallow.append(rset)

        if _txallow.range220_230 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range220_230)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range220_230", "220-230 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range220_230)
        txallow.append(rset)

        if _txallow.range230_240 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range230_240)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range230_240", "230-240 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range230_240)
        txallow.append(rset)

        if _txallow.range240_250 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range240_250)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range240_250", "240-250 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range240_250)
        txallow.append(rset)

        if _txallow.range250_260 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range250_260)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range250_260", "250-260 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range250_260)
        txallow.append(rset)

        if _txallow.range260_270 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range260_270)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range260_270", "260-270 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range260_270)
        txallow.append(rset)

        if _txallow.range270_280 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range270_280)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range270_280", "270-280 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range270_280)
        txallow.append(rset)

        if _txallow.range280_290 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range280_290)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range280_290", "280-290 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range280_290)
        txallow.append(rset)

        if _txallow.range290_300 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range290_300)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range290_300", "290-300 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range290_300)
        txallow.append(rset)

        if _txallow.range300_310 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range300_310)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range300_310", "300-310 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range300_310)
        txallow.append(rset)

        if _txallow.range310_320 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range310_320)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range310_320", "310-320 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range310_320)
        txallow.append(rset)

        if _txallow.range320_330 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range320_330)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range320_330", "320-330 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range320_330)
        txallow.append(rset)

        if _txallow.range330_340 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range330_340)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range330_340", "330-340 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range330_340)
        txallow.append(rset)

        if _txallow.range340_350 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range340_350)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range340_350", "340-350 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range340_350)
        txallow.append(rset)

        if _txallow.range350_360 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range350_360)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range350_360", "350-360 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range350_360)
        txallow.append(rset)

        if _txallow.range360_370 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range360_370)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range360_370", "360-370 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range360_370)
        txallow.append(rset)

        if _txallow.range370_380 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range370_380)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range370_380", "370-380 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range370_380)
        txallow.append(rset)

        if _txallow.range380_390 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range380_390)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range380_390", "380-390 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range380_390)
        txallow.append(rset)

        if _txallow.range390_400 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range390_400)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range390_400", "390-400 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range390_400)
        txallow.append(rset)

        if _txallow.range520_530 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range520_530)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range520_530", "520-530 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range520_530)
        txallow.append(rset)

        if _txallow.range530_540 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range530_540)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range530_540", "530-540 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range530_540)
        txallow.append(rset)

        if _txallow.range540_550 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range540_550)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range540_550", "540-550 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range540_550)
        txallow.append(rset)

        if _txallow.range550_560 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range550_560)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range550_560", "550-560 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range550_560)
        txallow.append(rset)

        if _txallow.range560_570 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range560_570)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range560_570", "560-570 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range560_570)
        txallow.append(rset)

        if _txallow.range570_580 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range570_580)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range570_580", "570-580 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range570_580)
        txallow.append(rset)

        if _txallow.range580_590 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range580_590)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range580_590", "580-590 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range580_590)
        txallow.append(rset)

        if _txallow.range590_600 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range590_600)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range590_600", "590-600 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range590_600)
        txallow.append(rset)

        if _txallow.range600_610 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range600_610)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range600_610", "600-610 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range600_610)
        txallow.append(rset)

        if _txallow.range610_620 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range610_620)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range610_620", "610-620 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range610_620)
        txallow.append(rset)

        if _txallow.range620_630 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range620_630)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range620_630", "620-630 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range620_630)
        txallow.append(rset)

        if _txallow.range630_640 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range630_640)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range630_640", "630-640 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range630_640)
        txallow.append(rset)

        if _txallow.range640_650 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range640_650)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range640_650", "640-650 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range640_650)
        txallow.append(rset)

        if _txallow.range650_660 in TXALLOW_VALUES:
            idx = TXALLOW_VALUES.index(_txallow.range650_660)
        else:
            idx = TXALLOW_VALUES.index(0xFF)
        rs = RadioSettingValueList(TXALLOW_CHOICES, TXALLOW_CHOICES[idx])
        rset = RadioSetting("txallow.range650_660", "650-660 MHz", rs)
        rset.set_apply_callback(apply_txallow_listvalue, _txallow.range650_660)
        txallow.append(rset)

        return top

    def set_settings(self, settings):
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    if "." in element.get_name():
                        bits = element.get_name().split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = self._memobj.settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("Using apply callback")
                        element.run_apply_callback()
                    elif setting == "dsrangea":
                        setattr(obj, setting, int(element.value) - 1)
                    elif setting == "dsrangeb":
                        setattr(obj, setting, int(element.value) - 1)
                    elif element.value.get_mutable():
                        LOG.debug("Setting %s = %s" % (setting, element.value))
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise

    @classmethod
    def match_model(cls, filedata, filename):
        # This radio has always been post-metadata, so never do
        # old-school detection
        return False


@directory.register
class AbbreeAR518Radio(IradioUV5118):
    """ABBREE AR518"""
    VENDOR = "Abbree"
    MODEL = "AR-518"



# Copyright 2010 Vernon Mauery <vernon@mauery.org>
# Copyright 2016 Angus Ainslie <angus@akkea.ca>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp import chirp_common, errors, directory
from chirp import bitwise, memmap
from chirp.drivers import kenwood_live
from chirp.settings import RadioSettingGroup, RadioSetting, RadioSettings
from chirp.settings import RadioSettingValueString
from chirp.settings import RadioSettingValueList
import time
import struct
import sys
import logging

LOG = logging.getLogger(__name__)

# TH-D72 memory map
# 0x0000..0x0200: startup password and other stuff
# 0x0200..0x0400: current channel and other settings
#   0x244,0x246: last menu numbers
#   0x249: last f menu number
# 0x0400..0x0c00: APRS settings and likely other settings
# 0x0c00..0x1500: memory channel flags
# 0x1500..0x5380: 0-999 channels
# 0x5380..0x54c0: 0-9 scan channels
# 0x54c0..0x5560: 0-9 wx channels
# 0x5560..0x5e00: ?
# 0x5e00..0x7d40: 0-999 channel names
# 0x7d40..0x7de0: ?
# 0x7de0..0x7e30: wx channel names
# 0x7e30..0x7ed0: ?
# 0x7ed0..0x7f20: group names
# 0x7f20..0x8b00: ?
# 0x8b00..0x9c00: last 20 APRS entries
# 0x9c00..0xe500: ?
# 0xe500..0xe7d0: startup bitmap
# 0xe7d0..0xe800: startup bitmap filename
# 0xe800..0xead0: gps-logger bitmap
# 0xe8d0..0xeb00: gps-logger bipmap filename
# 0xeb00..0xff00: ?
# 0xff00..0xffff: stuff?

# memory channel
# 0 1 2 3  4 5     6            7     8     9    a          b c d e   f
# [freq ]  ? mode  tmode/duplex rtone ctone dtcs cross_mode [offset]  ?

mem_format = """
#seekto 0x0000;
struct {
  ul16 version;
  u8   shouldbe32;
  u8   efs[11];
  u8   unknown0[3];
  u8   radio_custom_image;
  u8   gps_custom_image;
  u8   unknown1[7];
  u8   passwd[6];
} frontmatter;

#seekto 0x02c0;
struct {
  ul32 start_freq;
  ul32 end_freq;
} prog_vfo[6];

#seekto 0x0300;
struct {
  char power_on_msg[8];
  u8 unknown0[8];
  u8 unknown1[2];
  u8 lamp_timer;
  u8 contrast;
  u8 battery_saver;
  u8 APO;
  u8 unknown2;
  u8 key_beep;
  u8 unknown3[8];
  u8 unknown4;
  u8 balance;
  u8 unknown5[23];
  u8 lamp_control;
} settings;

#seekto 0x0c00;
struct {
  u8 disabled:4,
     prog_vfo:4;
  u8 skip;
} flag[1032];

#seekto 0x1500;
struct {
  ul32 freq;
  u8 unknown1:4,
     tune_step:4;
  u8 mode;
  u8 tone_mode:4,
     duplex:4;
  u8 rtone;
  u8 ctone;
  u8 dtcs;
  u8 cross_mode;
  ul32 offset;
  u8 unknown2:4,
     split_tune_step:4;
} memory[1032];

#seekto 0x5e00;
struct {
    char name[8];
} channel_name[1000];

#seekto 0x7de0;
struct {
    char name[8];
} wx_name[10];

#seekto 0x7ed0;
struct {
    char name[8];
} group_name[10];
"""

THD72_SPECIAL = {}

for i in range(0, 10):
    THD72_SPECIAL["L%i" % i] = 1000 + (i * 2)
    THD72_SPECIAL["U%i" % i] = 1000 + (i * 2) + 1
for i in range(0, 10):
    THD72_SPECIAL["WX%i" % (i + 1)] = 1020 + i
THD72_SPECIAL["C VHF"] = 1030
THD72_SPECIAL["C UHF"] = 1031

THD72_SPECIAL_REV = {}
for k, v in THD72_SPECIAL.items():
    THD72_SPECIAL_REV[v] = k

TMODES = {
    0x08: "Tone",
    0x04: "TSQL",
    0x02: "DTCS",
    0x01: "Cross",
    0x00: "",
}
TMODES_REV = {
    "": 0x00,
    "Cross": 0x01,
    "DTCS": 0x02,
    "TSQL": 0x04,
    "Tone": 0x08,
}

MODES = {
    0x00: "FM",
    0x01: "NFM",
    0x02: "AM",
}

MODES_REV = {
    "FM": 0x00,
    "NFM": 0x01,
    "AM": 0x2,
}

DUPLEX = {
    0x00: "",
    0x01: "+",
    0x02: "-",
    0x04: "split",
}
DUPLEX_REV = {
    "": 0x00,
    "+": 0x01,
    "-": 0x02,
    "split": 0x04,
}

TUNE_STEPS = [5.0, 6.25, 8.33, 10.0, 12.5, 15.0, 20.0, 25.0, 30.0, 50.0, 100.0]

EXCH_R = "R\x00\x00\x00\x00"
EXCH_W = "W\x00\x00\x00\x00"

DEFAULT_PROG_VFO = (
    (136000000, 174000000),
    (410000000, 470000000),
    (118000000, 136000000),
    (136000000, 174000000),
    (320000000, 400000000),
    (400000000, 524000000),
)

D72_FILE_HEADER = (
    b'MCP-4A\xFF\xFFV1.04\xFF\xFF\xFF' +
    b'TH-D72' + (b'\xFF' * 10) +
    b'\x31' + (b'\xFF' * 15) +
    b'\xFF' * (5 * 16) +
    b'AMB0' + (b'\xFF' * 12) +
    b'\xFF' * (7 * 16))


def get_prog_vfo(frequency):
    for i, (start, end) in enumerate(DEFAULT_PROG_VFO):
        if start <= frequency < end:
            return i
    raise ValueError("Frequency is out of range.")


@directory.register
class THD72Radio(chirp_common.CloneModeRadio):

    BAUD_RATE = 9600
    VENDOR = "Kenwood"
    MODEL = "TH-D72 (clone mode)"
    HARDWARE_FLOW = sys.platform == "darwin"  # only OS X driver needs hw flow
    NEEDS_COMPAT_SERIAL = False
    FORMATS = [directory.register_format('Kenwood MCP4A', '*.mc4')]

    mem_upper_limit = 1022
    _memsize = 65536
    _model = ""  # FIXME: REMOVE
    _dirty_blocks = []

    _LCD_CONTRAST = ["Level %d" % x for x in range(1, 16)]
    _LAMP_CONTROL = ["Manual", "Auto"]
    _LAMP_TIMER = ["Seconds %d" % x for x in range(2, 11)]
    _BATTERY_SAVER = ["OFF", "0.03 Seconds", "0.2 Seconds", "0.4 Seconds",
                      "0.6 Seconds", "0.8 Seconds", "1 Seconds", "2 Seconds",
                      "3 Seconds", "4 Seconds", "5 Seconds"]
    _APO = ["OFF", "15 Minutes", "30 Minutes", "60 Minutes"]
    _AUDIO_BALANCE = ["Center", "A +50%", "A +100%", "B +50%", "B +100%"]
    _KEY_BEEP = ["OFF", "Radio & GPS", "Radio Only", "GPS Only"]

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.memory_bounds = (0, 999)
        rf.valid_bands = [(118000000, 174000000),
                          (320000000, 524000000)]
        rf.has_cross = True
        rf.can_odd_split = True
        rf.has_dtcs_polarity = False
        rf.has_tuning_step = True
        rf.has_bank = False
        rf.has_settings = True
        rf.valid_tuning_steps = []
        rf.valid_modes = list(MODES_REV.keys())
        rf.valid_tmodes = list(TMODES_REV.keys())
        rf.valid_duplexes = list(DUPLEX_REV.keys())
        rf.valid_tuning_steps = list(TUNE_STEPS)
        rf.valid_tones = list(kenwood_live.KENWOOD_TONES)
        rf.valid_skips = ["", "S"]
        rf.valid_characters = chirp_common.CHARSET_ALPHANUMERIC
        rf.valid_name_length = 8
        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(mem_format, self._mmap)
        self._dirty_blocks = []

    def _detect_baud(self):
        for baud in [9600, 19200, 38400, 57600]:
            self.pipe.baudrate = baud
            try:
                self.pipe.write(b"\r\r")
            except:
                break
            self.pipe.read(32)
            try:
                id = self.get_id()
                LOG.info("Radio %s at %i baud" % (id, baud))
                return True
            except errors.RadioError:
                pass

        raise errors.RadioError("No response from radio")

    def get_special_locations(self):
        return sorted(THD72_SPECIAL.keys())

    def add_dirty_block(self, memobj):
        block = memobj._offset // 256
        if block not in self._dirty_blocks:
            self._dirty_blocks.append(block)
        self._dirty_blocks.sort()
        print("dirty blocks: ", self._dirty_blocks)

    def get_channel_name(self, number):
        if number < 999:
            name = str(self._memobj.channel_name[number].name) + '\xff'
        elif number >= 1020 and number < 1030:
            number -= 1020
            name = str(self._memobj.wx_name[number].name) + '\xff'
        else:
            return ''
        return name[:name.index('\xff')].rstrip()

    def set_channel_name(self, number, name):
        name = name[:8] + '\xff' * 8
        if number < 999:
            self._memobj.channel_name[number].name = name[:8]
            self.add_dirty_block(self._memobj.channel_name[number])
        elif number >= 1020 and number < 1030:
            number -= 1020
            self._memobj.wx_name[number].name = name[:8]
            self.add_dirty_block(self._memobj.wx_name[number])

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number]) + \
            repr(self._memobj.flag[number])

    def get_memory(self, number):
        if isinstance(number, str):
            try:
                number = THD72_SPECIAL[number]
            except KeyError:
                raise errors.InvalidMemoryLocation("Unknown channel %s" %
                                                   number)

        if number < 0 or number > (max(THD72_SPECIAL.values()) + 1):
            raise errors.InvalidMemoryLocation(
                "Number must be between 0 and 999")

        _mem = self._memobj.memory[number]
        flag = self._memobj.flag[number]

        mem = chirp_common.Memory()
        mem.number = number

        if number > 999:
            mem.extd_number = THD72_SPECIAL_REV[number]
        if flag.disabled == 0xf:
            mem.empty = True
            return mem

        mem.name = self.get_channel_name(number)
        mem.freq = int(_mem.freq)
        mem.tmode = TMODES[int(_mem.tone_mode)]
        mem.rtone = kenwood_live.KENWOOD_TONES[_mem.rtone]
        mem.ctone = kenwood_live.KENWOOD_TONES[_mem.ctone]
        mem.dtcs = chirp_common.DTCS_CODES[_mem.dtcs]
        mem.duplex = DUPLEX[int(_mem.duplex)]
        mem.offset = int(_mem.offset)
        mem.mode = MODES[int(_mem.mode)]
        mem.tuning_step = TUNE_STEPS[int(_mem.tune_step)]

        if number < 999:
            mem.skip = chirp_common.SKIP_VALUES[int(flag.skip)]
            mem.cross_mode = chirp_common.CROSS_MODES[_mem.cross_mode]
        if number > 999:
            mem.cross_mode = chirp_common.CROSS_MODES[0]
            mem.immutable = ["number", "bank", "extd_number", "cross_mode"]
            if number >= 1020 and number < 1030:
                mem.immutable += ["freq", "offset", "tone", "mode",
                                  "tmode", "ctone", "skip"]  # FIXME: ALL
            else:
                mem.immutable += ["name"]

        return mem

    def set_memory(self, mem):
        LOG.debug("set_memory(%d)" % mem.number)
        if mem.number < 0 or mem.number > (max(THD72_SPECIAL.values()) + 1):
            raise errors.InvalidMemoryLocation(
                "Number must be between 0 and 999")

        # weather channels can only change name, nothing else
        if mem.number >= 1020 and mem.number < 1030:
            self.set_channel_name(mem.number, mem.name)
            return

        flag = self._memobj.flag[mem.number]
        self.add_dirty_block(self._memobj.flag[mem.number])

        # only delete non-WX channels
        was_empty = flag.disabled == 0xf
        if mem.empty:
            flag.disabled = 0xf
            return
        flag.disabled = 0

        _mem = self._memobj.memory[mem.number]
        self.add_dirty_block(_mem)
        if was_empty:
            self.initialize(_mem)

        _mem.freq = mem.freq

        if mem.number < 999:
            self.set_channel_name(mem.number, mem.name)

        _mem.tone_mode = TMODES_REV[mem.tmode]
        _mem.rtone = kenwood_live.KENWOOD_TONES.index(mem.rtone)
        _mem.ctone = kenwood_live.KENWOOD_TONES.index(mem.ctone)
        _mem.dtcs = chirp_common.DTCS_CODES.index(mem.dtcs)
        _mem.cross_mode = chirp_common.CROSS_MODES.index(mem.cross_mode)
        _mem.duplex = DUPLEX_REV[mem.duplex]
        _mem.offset = mem.offset
        _mem.mode = MODES_REV[mem.mode]
        _mem.tune_step = TUNE_STEPS.index(mem.tuning_step)

        if mem.duplex == 'split':
            _mem.split_tune_step = TUNE_STEPS.index(
                chirp_common.required_step(mem.offset))
        else:
            _mem.split_tune_step = _mem.tune_step

        prog_vfo = get_prog_vfo(mem.freq)
        flag.prog_vfo = prog_vfo

        if mem.number < 999:
            flag.skip = chirp_common.SKIP_VALUES.index(mem.skip)

    def sync_in(self):
        self._detect_baud()
        self._mmap = self.download()
        self.process_mmap()

    def sync_out(self):
        self._detect_baud()
        if len(self._dirty_blocks):
            self.upload(self._dirty_blocks)
        else:
            self.upload()

    def read_block(self, block, count=256):
        self.pipe.write(struct.pack("<cBHB", b"R", 0, block, 0))
        r = self.pipe.read(5)
        if len(r) != 5:
            raise Exception("Did not receive block response")

        cmd, _zero, _block, zero = struct.unpack("<cBHB", r)
        if cmd != b"W" or _block != block:
            raise Exception("Invalid response: %s %i" % (cmd, _block))

        data = b""
        while len(data) < count:
            data += self.pipe.read(count - len(data))

        self.pipe.write(bytes([0x06]))
        if self.pipe.read(1) != bytes([0x06]):
            raise Exception("Did not receive post-block ACK!")

        return data

    def write_block(self, block, map):
        self.pipe.write(struct.pack("<cBHB", b"W", 0, block, 0))
        base = block * 256
        self.pipe.write(map[base:base + 256])

        ack = self.pipe.read(1)

        return ack == bytes([0x06])

    def download(self, raw=False, blocks=None):
        if blocks is None:
            blocks = range(self._memsize // 256)
        else:
            blocks = [b for b in blocks if b < self._memsize // 256]

        if self.command("0M PROGRAM") != "0M":
            raise errors.RadioError("No response from self")

        allblocks = range(self._memsize // 256)
        self.pipe.baudrate = 57600
        try:
            self.pipe.setRTS()
        except AttributeError:
            self.pipe.rts = True
        self.pipe.read(1)
        data = b""
        LOG.debug("reading blocks %d..%d" % (blocks[0], blocks[-1]))
        total = len(blocks)
        count = 0
        for i in allblocks:
            if i not in blocks:
                data += 256 * b'\xff'
                continue
            data += self.read_block(i)
            count += 1
            if self.status_fn:
                s = chirp_common.Status()
                s.msg = "Cloning from radio"
                s.max = total
                s.cur = count
                self.status_fn(s)

        self.pipe.write(b"E")

        if raw:
            return data
        return memmap.MemoryMapBytes(data)

    def upload(self, blocks=None):
        if blocks is None:
            blocks = range((self._memsize // 256) - 2)
        else:
            blocks = [b for b in blocks if b < self._memsize // 256]

        if self.command("0M PROGRAM") != "0M":
            raise errors.RadioError("No response from self")

        self.pipe.baudrate = 57600
        try:
            self.pipe.setRTS()
        except AttributeError:
            self.pipe.rts = True
        self.pipe.read(1)
        LOG.debug("writing blocks %d..%d" % (blocks[0], blocks[-1]))
        total = len(blocks)
        count = 0
        for i in blocks:
            r = self.write_block(i, self._mmap)
            count += 1
            if not r:
                raise errors.RadioError("self NAK'd block %i" % i)
            if self.status_fn:
                s = chirp_common.Status()
                s.msg = "Cloning to radio"
                s.max = total
                s.cur = count
                self.status_fn(s)

        self.pipe.write(b"E")
        # clear out blocks we uploaded from the dirty blocks list
        self._dirty_blocks = [b for b in self._dirty_blocks if b not in blocks]

    def command(self, cmd, timeout=0.5):
        start = time.time()

        data = b""
        LOG.debug("PC->D72: %s" % cmd)
        self.pipe.write((cmd + "\r").encode())
        while not data.endswith(b"\r") and (time.time() - start) < timeout:
            data += self.pipe.read(1)
        LOG.debug("D72->PC: %s" % data.strip())
        return data.decode().strip()

    def get_id(self):
        r = self.command("ID")
        if r.startswith("ID "):
            return r.split(" ")[1]
        else:
            raise errors.RadioError("No response to ID command")

    def initialize(self, mmap):
        mmap.set_raw("\x00\xc8\xb3\x08\x00\x01\x00\x08"
                     "\x08\x00\xc0\x27\x09\x00\x00\x00")

    def _get_settings(self):
        top = RadioSettings(self._get_display_settings(),
                            self._get_audio_settings(),
                            self._get_battery_settings())
        return top

    def set_settings(self, settings):
        _mem = self._memobj
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            if not element.changed():
                continue
            try:
                if element.has_apply_callback():
                    LOG.debug("Using apply callback")
                    try:
                        element.run_apply_callback()
                    except NotImplementedError as e:
                        LOG.error("thd72: %s", e)
                    continue

                # Find the object containing setting.
                obj = _mem
                bits = element.get_name().split(".")
                setting = bits[-1]
                for name in bits[:-1]:
                    if name.endswith("]"):
                        name, index = name.split("[")
                        index = int(index[:-1])
                        obj = getattr(obj, name)[index]
                    else:
                        obj = getattr(obj, name)

                try:
                    old_val = getattr(obj, setting)
                    LOG.debug("Setting %s(%r) <= %s" % (
                        element.get_name(), old_val, element.value))
                    setattr(obj, setting, element.value)
                except AttributeError as e:
                    LOG.error("Setting %s is not in the memory map: %s" %
                              (element.get_name(), e))
            except Exception:
                LOG.debug(element.get_name())
                raise

    def get_settings(self):
        try:
            return self._get_settings()
        except:
            import traceback
            LOG.error("Failed to parse settings: %s", traceback.format_exc())
            return None

    @classmethod
    def apply_power_on_msg(cls, setting, obj):
        message = setting.value.get_value()
        setattr(obj, "power_on_msg", cls._add_ff_pad(message, 8))

    def apply_lcd_contrast(cls, setting, obj):
        rawval = setting.value.get_value()
        val = cls._LCD_CONTRAST.index(rawval) + 1
        obj.contrast = val

    def apply_lamp_control(cls, setting, obj):
        rawval = setting.value.get_value()
        val = cls._LAMP_CONTROL.index(rawval)
        obj.lamp_control = val

    def apply_lamp_timer(cls, setting, obj):
        rawval = setting.value.get_value()
        val = cls._LAMP_TIMER.index(rawval) + 2
        obj.lamp_timer = val

    def _get_display_settings(self):
        menu = RadioSettingGroup("display", "Display")
        display_settings = self._memobj.settings

        val = RadioSettingValueString(
            0, 8, str(display_settings.power_on_msg).rstrip("\xFF"))
        rs = RadioSetting("display.power_on_msg", "Power on message", val)
        rs.set_apply_callback(self.apply_power_on_msg, display_settings)
        menu.append(rs)

        val = RadioSettingValueList(
            self._LCD_CONTRAST,
            self._LCD_CONTRAST[display_settings.contrast - 1])
        rs = RadioSetting("display.contrast", "LCD Contrast",
                          val)
        rs.set_apply_callback(self.apply_lcd_contrast, display_settings)
        menu.append(rs)

        val = RadioSettingValueList(
            self._LAMP_CONTROL,
            self._LAMP_CONTROL[display_settings.lamp_control])
        rs = RadioSetting("display.lamp_control", "Lamp Control",
                          val)
        rs.set_apply_callback(self.apply_lamp_control, display_settings)
        menu.append(rs)

        val = RadioSettingValueList(
            self._LAMP_TIMER,
            self._LAMP_TIMER[display_settings.lamp_timer - 2])
        rs = RadioSetting("display.lamp_timer", "Lamp Timer",
                          val)
        rs.set_apply_callback(self.apply_lamp_timer, display_settings)
        menu.append(rs)

        return menu

    def apply_battery_saver(cls, setting, obj):
        rawval = setting.value.get_value()
        val = cls._BATTERY_SAVER.index(rawval)
        obj.battery_saver = val

    def apply_APO(cls, setting, obj):
        rawval = setting.value.get_value()
        val = cls._APO.index(rawval)
        obj.APO = val

    def _get_battery_settings(self):
        menu = RadioSettingGroup("battery", "Battery")
        battery_settings = self._memobj.settings

        val = RadioSettingValueList(
            self._BATTERY_SAVER,
            self._BATTERY_SAVER[battery_settings.battery_saver])
        rs = RadioSetting("battery.battery_saver", "Battery Saver",
                          val)
        rs.set_apply_callback(self.apply_battery_saver, battery_settings)
        menu.append(rs)

        val = RadioSettingValueList(
            self._APO,
            self._APO[battery_settings.APO])
        rs = RadioSetting("battery.APO", "Auto Power Off",
                          val)
        rs.set_apply_callback(self.apply_APO, battery_settings)
        menu.append(rs)

        return menu

    def apply_balance(cls, setting, obj):
        rawval = setting.value.get_value()
        val = cls._AUDIO_BALANCE.index(rawval)
        obj.balance = val

    def apply_key_beep(cls, setting, obj):
        rawval = setting.value.get_value()
        val = cls._KEY_BEEP.index(rawval)
        obj.key_beep = val

    def _get_audio_settings(self):
        menu = RadioSettingGroup("audio", "Audio")
        audio_settings = self._memobj.settings

        val = RadioSettingValueList(
            self._AUDIO_BALANCE,
            self._AUDIO_BALANCE[audio_settings.balance])
        rs = RadioSetting("audio.balance", "Balance",
                          val)
        rs.set_apply_callback(self.apply_balance, audio_settings)
        menu.append(rs)

        val = RadioSettingValueList(
            self._KEY_BEEP,
            self._KEY_BEEP[audio_settings.key_beep])
        rs = RadioSetting("audio.key_beep", "Key Beep",
                          val)
        rs.set_apply_callback(self.apply_key_beep, audio_settings)
        menu.append(rs)

        return menu

    @staticmethod
    def _add_ff_pad(val, length):
        return val.ljust(length, "\xFF")[:length]

    @classmethod
    def _strip_ff_pads(cls, messages):
        result = []
        for msg_text in messages:
            result.append(str(msg_text).rstrip("\xFF"))
        return result

    def load_mmap(self, filename):
        if filename.lower().endswith('.mc4'):
            with open(filename, 'rb') as f:
                f.seek(0x100)
                self._mmap = memmap.MemoryMapBytes(f.read())
                LOG.info('Loaded MCP file at offset 0x100')
            self.process_mmap()
        else:
            chirp_common.CloneModeRadio.load_mmap(self, filename)

    def save_mmap(self, filename):
        if filename.lower().endswith('.mc4'):
            with open(filename, 'wb') as f:
                f.write(D72_FILE_HEADER)
                f.write(self._mmap.get_packed())
                LOG.info('Wrote MCP file')
        else:
            chirp_common.CloneModeRadio.save_mmap(self, filename)

    @classmethod
    def match_model(cls, filedata, filename):
        if filename.endswith('.mc4'):
            return True
        else:
            return super(THD72Radio, cls).match_model(filedata, filename)



from chirp.drivers import __init__
from chirp.drivers import bj9900
from chirp.drivers import anytone
from chirp.drivers import anytone778uv
from chirp.drivers import anytone_ht
from chirp.drivers import anytone_iii
from chirp.drivers import ap510
from chirp.drivers import baofeng_common
from chirp.drivers import baofeng_uv3r
from chirp.drivers import baofeng_wp970i
from chirp.drivers import bjuv55
from chirp.drivers import boblov_x3plus
from chirp.drivers import btech
from chirp.drivers import fake
from chirp.drivers import fd268
from chirp.drivers import ft1500m
from chirp.drivers import ft1802
from chirp.drivers import ft1d
from chirp.drivers import ft2900
from chirp.drivers import ft2d
from chirp.drivers import ft4
from chirp.drivers import ft2800
from chirp.drivers import ft450d
from chirp.drivers import ft50
from chirp.drivers import ft60
from chirp.drivers import ft70
from chirp.drivers import ft7100
from chirp.drivers import ft7800
from chirp.drivers import ft8100
from chirp.drivers import ft817
from chirp.drivers import ft818
from chirp.drivers import ft857
from chirp.drivers import ft90
from chirp.drivers import ftlx011
from chirp.drivers import ftm3200d
from chirp.drivers import ftm350
from chirp.drivers import ftm7250d
from chirp.drivers import ga510
from chirp.drivers import generic_csv
from chirp.drivers import generic_tpe
from chirp.drivers import gmrsuv1
from chirp.drivers import gmrsv2
from chirp.drivers import h777
from chirp.drivers import hf90
from chirp.drivers import bf_t1
from chirp.drivers import ic208
from chirp.drivers import hg_uv98
from chirp.drivers import hobbypcb
from chirp.drivers import ic2300
from chirp.drivers import ic2100
from chirp.drivers import ic2200
from chirp.drivers import ic2720
from chirp.drivers import ic2730
from chirp.drivers import ic9x_icf
from chirp.drivers import ic2820
from chirp.drivers import ic9x
from chirp.drivers import ic9x_ll
from chirp.drivers import ic9x_icf_ll
from chirp.drivers import icf
from chirp.drivers import icomciv
from chirp.drivers import icp7
from chirp.drivers import ict8
from chirp.drivers import icq7
from chirp.drivers import ict70
from chirp.drivers import ict7h
from chirp.drivers import icv80
from chirp.drivers import id31
from chirp.drivers import icv86
from chirp.drivers import icw32
from chirp.drivers import icx8x
from chirp.drivers import icx90
from chirp.drivers import id5100
from chirp.drivers import id51
from chirp.drivers import id51plus
from chirp.drivers import id880
from chirp.drivers import id800
from chirp.drivers import idrp
from chirp.drivers import kg935g
from chirp.drivers import iradio_uv_5118
from chirp.drivers import kenwood_hmk
from chirp.drivers import kenwood_itm
from chirp.drivers import kenwood_live
from chirp.drivers import kguv8d
from chirp.drivers import kguv980p
from chirp.drivers import kguv8dplus
from chirp.drivers import kguv8e
from chirp.drivers import kguv920pa
from chirp.drivers import kguv9dplus
from chirp.drivers import leixen
from chirp.drivers import kyd
from chirp.drivers import kyd_IP620
from chirp.drivers import lt725uv
from chirp.drivers import mursv1
from chirp.drivers import alinco
from chirp.drivers import puxing
from chirp.drivers import puxing_px888k
from chirp.drivers import radioddity_r2
from chirp.drivers import radtel_t18
from chirp.drivers import retevis_rb15
from chirp.drivers import retevis_rb17p
from chirp.drivers import retevis_rb28
from chirp.drivers import retevis_rt1
from chirp.drivers import retevis_rt21
from chirp.drivers import retevis_rt22
from chirp.drivers import retevis_rt23
from chirp.drivers import retevis_rt26
from chirp.drivers import retevis_rt76p
from chirp.drivers import retevis_rt87
from chirp.drivers import retevis_rt98
from chirp.drivers import bf_t8
from chirp.drivers import rh5r_v2
from chirp.drivers import tdxone_tdq8a
from chirp.drivers import template
from chirp.drivers import tg_uv2p
from chirp.drivers import th350
from chirp.drivers import th7800
from chirp.drivers import th9000
from chirp.drivers import th9800
from chirp.drivers import th_uv3r
from chirp.drivers import th_uv3r25
from chirp.drivers import th_uv8000
from chirp.drivers import th_uv88
from chirp.drivers import th_uvf8d
from chirp.drivers import thd72
from chirp.drivers import thd74
from chirp.drivers import thuv1f
from chirp.drivers import tk270
from chirp.drivers import tk3140
from chirp.drivers import tk760
from chirp.drivers import tk760g
from chirp.drivers import tk8102
from chirp.drivers import tk8160
from chirp.drivers import tk8180
from chirp.drivers import tmd710
from chirp.drivers import tmv71
from chirp.drivers import tmv71_ll
from chirp.drivers import ts2000
from chirp.drivers import ts480
from chirp.drivers import ts590
from chirp.drivers import ts850
from chirp.drivers import uv5r
from chirp.drivers import uv5x3
from chirp.drivers import uv6r
from chirp.drivers import uvb5
from chirp.drivers import vgc
from chirp.drivers import vx170
from chirp.drivers import vx2
from chirp.drivers import vx3
from chirp.drivers import vx5
from chirp.drivers import vx510
from chirp.drivers import vx6
from chirp.drivers import vx7
from chirp.drivers import vx8
from chirp.drivers import vxa700
from chirp.drivers import wouxun
from chirp.drivers import wouxun_common
from chirp.drivers import kenwood_d7
from chirp.drivers import yaesu_clone
from chirp.drivers import tdh8
from chirp.drivers import radtel_rt490
from chirp.drivers import baofeng_uv17Pro
from chirp.drivers import uvk5
from chirp.drivers import baofeng_uv17
from chirp.drivers import iradio_uv_5118plus
from chirp.drivers import mml_jc8810
from chirp.drivers import uvk5_egzumer
__all__ = ['__init__', 'bj9900', 'anytone', 'anytone778uv', 'anytone_ht', 'anytone_iii', 'ap510', 'baofeng_common', 'baofeng_uv3r', 'baofeng_wp970i', 'bjuv55', 'boblov_x3plus', 'btech', 'fake', 'fd268', 'ft1500m', 'ft1802', 'ft1d', 'ft2900', 'ft2d', 'ft4', 'ft2800', 'ft450d', 'ft50', 'ft60', 'ft70', 'ft7100', 'ft7800', 'ft8100', 'ft817', 'ft818', 'ft857', 'ft90', 'ftlx011', 'ftm3200d', 'ftm350', 'ftm7250d', 'ga510', 'generic_csv', 'generic_tpe', 'gmrsuv1', 'gmrsv2', 'h777', 'hf90', 'bf_t1', 'ic208', 'hg_uv98', 'hobbypcb', 'ic2300', 'ic2100', 'ic2200', 'ic2720', 'ic2730', 'ic9x_icf', 'ic2820', 'ic9x', 'ic9x_ll', 'ic9x_icf_ll', 'icf', 'icomciv', 'icp7', 'ict8', 'icq7', 'ict70', 'ict7h', 'icv80', 'id31', 'icv86', 'icw32', 'icx8x', 'icx90', 'id5100', 'id51', 'id51plus', 'id880', 'id800', 'idrp', 'kg935g', 'iradio_uv_5118', 'kenwood_hmk', 'kenwood_itm', 'kenwood_live', 'kguv8d', 'kguv980p', 'kguv8dplus', 'kguv8e', 'kguv920pa', 'kguv9dplus', 'leixen', 'kyd', 'kyd_IP620', 'lt725uv', 'mursv1', 'alinco', 'puxing', 'puxing_px888k', 'radioddity_r2', 'radtel_t18', 'retevis_rb15', 'retevis_rb17p', 'retevis_rb28', 'retevis_rt1', 'retevis_rt21', 'retevis_rt22', 'retevis_rt23', 'retevis_rt26', 'retevis_rt76p', 'retevis_rt87', 'retevis_rt98', 'bf_t8', 'rh5r_v2', 'tdxone_tdq8a', 'template', 'tg_uv2p', 'th350', 'th7800', 'th9000', 'th9800', 'th_uv3r', 'th_uv3r25', 'th_uv8000', 'th_uv88', 'th_uvf8d', 'thd72', 'thd74', 'thuv1f', 'tk270', 'tk3140', 'tk760', 'tk760g', 'tk8102', 'tk8160', 'tk8180', 'tmd710', 'tmv71', 'tmv71_ll', 'ts2000', 'ts480', 'ts590', 'ts850', 'uv5r', 'uv5x3', 'uv6r', 'uvb5', 'vgc', 'vx170', 'vx2', 'vx3', 'vx5', 'vx510', 'vx6', 'vx7', 'vx8', 'vxa700', 'wouxun', 'wouxun_common', 'kenwood_d7', 'yaesu_clone', 'tdh8', 'radtel_rt490', 'baofeng_uv17Pro', 'uvk5', 'baofeng_uv17', 'iradio_uv_5118plus', 'mml_jc8810', 'uvk5_egzumer']



# Copyright 2023 Dan Smith <chirp@f.danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import functools
import logging
import struct

from chirp import bitwise
from chirp import chirp_common
from chirp import directory
from chirp.drivers import tk8180
from chirp import errors
from chirp import memmap
from chirp import settings
from chirp import util

LOG = logging.getLogger(__name__)

mem_format = """
// Bitmask for zone in use
#seekto 0x0020;
u8 zoneflag[16];

// (identical?) bitmask for index spot in use
#seekto 0x0030;
u8 flag1[16];
#seekto 0x0050;
u8 flag2[16];

// This is filled in order, regardless of the ordering of the memories or
// their positions
#seekto 0x00E0;
struct {
    u8 zone;   // Zone number
    u8 memory; // Channel number
    u8 zero;
    u8 slot;  // Index into the memory array
} index[128];

#seekto 0x02E0;
struct {
    u8 used:1,
       unknown1:4, // Always 0x71?
       zoneaudiomode:2, // 00=QT/DQT 01=QT/DQT+OptSig 02 QT/DQTorOptSig
       optsigdecode:1;  // 0=Carrier 1=QT/DQT
    u8 count; // Number of memories in the zone
} zoneinfo[128];

#seekto 0x03E0;
struct {
    u8 number;
    u8 zone;
    lbcd rx_freq[4];
    lbcd tx_freq[4];
    u8 unknown1[2];
    ul16 rxtone;
    ul16 txtone;
    // BCL 0:no 1: carrier
    u8 unknown2_0:4,
       bcl:2, // 0:no 1:carrier 2:tone 3:optsig
       unknown2_01:1,
       wide:1;
    u8 beatshift:1,
       unknown2_1:1,
       highpower:1,
       scanadd:1,
       unknown2_2:4;
    u8 unknown2_3:6,
       pttid:2; // 0:off 1:BOT 2:EOT 3:Both
    char name[8];
    u8 unknown3_0:7,
       compander:1;
    u8 unknown3_1:7,
       pttidmute:1;
    u8 optsig:3, // 0:none 1:DTMF 2:2tone1 3:2tone2 4:2tone3 5:FleetSync
       unknown4:5;
    u8 unknown5[2];
} memories[128];

#seekto 0x14F0;
struct {
    char message1[32];
    char message2[32];
} messages;

#seekto 0x1730;
struct {
  char model[5];
  u8 variant;
} model;

// 28=volup 2A=None 15=2-tone 8=AUX 16=Call1 17=Call2 13=ChDn 12=ChUp
// 0B=DisplayCh 09=HornAlert 03=KeyLock 14=Brightness 02=Monitor 25=MonMom
// 0C=OST 19=PA 1A=SquelchLevel 05=Scan 06=ScanAdd 0A=Scrambler 18=Selcall
// 27=SendGPS 04=TA 28=VolUp 29=VolDn 00=ZoneUp 24=SqlOff 26=SqlOffMom
// 1C=Status 1D=SelCall+Status 01=ZoneDn
#seekto 0x1820;
struct {
  u8 skey;
  u8 akey;
  u8 bkey;
  u8 ckey;
  u8 triangle;
  u8 rightup;
  u8 rightdn;
  u8 leftup;
  u8 leftdn;
} keys;
"""

_KEYS = [
    'Zone Up',
    'Zone Down',
    'Monitor',
    'Key Lock',
    'Talk Around',
    'Scan',
    'Scan Add',
    '?',
    'AUX',
    'Horn Alert',
    'Scrambler',
    'Display Character',
    'OST',
    '?',  # 0D
    '?',  # 0E
    '?',  # 0F
    '?',  # 10
    '?',  # 11
    'Channel Up',
    'Channel Down',
    'Brightness',
    '2-Tone',
    'Call 1',
    'Call 2',
    'Selcall',
    'Public Address',
    'Squelch Level',
    '?',
    'Status',
    'Selcall+Status',
    '?',  # 1E
    '?',  # 1F
    '?',  # 20
    '?',  # 21
    '?',  # 22
    '?',  # 23
    'Squelch Off',
    'Monitor Momentary',
    'Squelch Off Momentary',
    'Send the GPS',
    'Volume Up',
    'Volume Down',
    'None',
    ]

KEYS = {i: key for i, key in enumerate(_KEYS)
        if not key.startswith('?')}

KEY_NAMES = {
    'A': 'akey',
    'B': 'bkey',
    'C': 'ckey',
    'S': 'skey',
    'Triangle': 'triangle',
    'Right Up': 'rightup',
    'Right Down': 'rightdn',
    'Left Up': 'leftup',
    'Left Down': 'leftdn',
}

POWER_LEVELS = [chirp_common.PowerLevel('Low', watts=25),
                chirp_common.PowerLevel('High', watts=50)]


def ends_program(fn):
    @functools.wraps(fn)
    def wrapper(radio):
        try:
            return fn(radio)
        finally:
            radio.pipe.write(b'E')
            ack = radio.pipe.read(1)
            if ack != b'\x06':
                LOG.warning('Radio did not ACK exiting program mode')

    return wrapper


@ends_program
def do_download(radio):
    tk8180.do_ident(radio)

    data = b''

    def status():
        status = chirp_common.Status()
        status.cur = len(data)
        status.max = radio._memsize
        status.msg = 'Cloning from radio'
        radio.status_fn(status)
        LOG.debug('Radio address 0x%04x' % len(data))

    # This seems to be "give me the whole thing"
    LOG.debug('Requesting clone image')
    radio.pipe.write(b'hRDA')
    ack = radio.pipe.read(1)
    if ack != b'\x06':
        LOG.error('No ack from radio; got %r' % ack)
        raise errors.RadioError('Radio refused to send memory')

    while len(data) < radio._memsize:
        cmd = radio.pipe.read(1)
        if cmd == b'E':
            LOG.debug('Got clone end')
            break
        elif cmd not in b'ZW':
            LOG.error('Unsupported command %r' % cmd)
            raise errors.RadioError('Radio sent unknown message')

        hdr = radio.pipe.read(3)
        addr, length = struct.unpack('>HB', hdr)
        LOG.debug('Radio: %r %04x %02x' % (cmd, addr, length))
        if cmd == b'Z':
            data += b'\xFF' * length
        elif cmd == b'W':
            block = radio.pipe.read(length)
            data += block
        radio.pipe.write(b'\x06')
        status()

    return memmap.MemoryMapBytes(data)


@ends_program
def do_upload(radio):
    tk8180.do_ident(radio)

    data = radio._mmap.get_packed()

    def status(addr):
        status = chirp_common.Status()
        status.cur = addr
        status.max = radio._memsize
        status.msg = "Cloning to radio"
        radio.status_fn(status)
        LOG.debug('Radio address 0x%04x' % addr)

    for addr in range(0, 0x30E0, 0x40):
        block = data[addr:addr + 0x40]
        hdr = struct.pack('>cHB', b'W', addr, 0x40)
        radio.pipe.write(hdr + block)
        ack = radio.pipe.read(1)
        if ack != b'\x06':
            LOG.error('Expected ack, got %r' % ack)
            raise errors.RadioError('Radio did not ACK block')
        status(addr)


class TKx160Radio(chirp_common.CloneModeRadio):
    VENDOR = 'Kenwood'
    FORMATS = [directory.register_format('Kenwood KPG-99D', '*.dat')]
    NEEDS_COMPAT_SERIAL = False
    _memsize = 0x4000

    def sync_in(self):
        try:
            self._mmap = do_download(self)
        except errors.RadioError:
            raise
        except Exception as e:
            LOG.exception('Failed download: %s' % e)
            raise errors.RadioError('Failed to communicate with radio')
        self.process_mmap()

    def sync_out(self):
        try:
            self._mmap = do_upload(self)
        except errors.RadioError:
            raise
        except Exception as e:
            LOG.exception('Failed upload: %s' % e)
            raise errors.RadioError('Failed to communicate with radio')

    def process_mmap(self):
        self._memobj = bitwise.parse(mem_format, self._mmap)

    @classmethod
    def match_model(cls, filedata, filename):
        return (filename.endswith('.dat') and
                filedata[0x1740:0x1746] == cls._model)

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.memory_bounds = (1, 128)
        rf.has_ctone = True
        rf.has_cross = True
        rf.has_bank = False
        rf.has_sub_devices = not self.VARIANT
        rf.has_tuning_step = False
        rf.has_rx_dtcs = True
        rf.has_settings = True
        rf.can_odd_split = True
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'DTCS', 'Cross']
        rf.valid_cross_modes = ['Tone->Tone', 'DTCS->', '->DTCS', "Tone->DTCS",
                                'DTCS->Tone', '->Tone', 'DTCS->DTCS']
        rf.valid_duplexes = ['', '-', '+', 'split', 'off']
        rf.valid_modes = ['FM', 'NFM']
        rf.valid_tuning_steps = [2.5, 5.0, 6.25, 12.5, 10.0, 15.0, 20.0,
                                 25.0, 50.0, 100.0]
        rf.valid_bands = self.VALID_BANDS
        rf.valid_characters = chirp_common.CHARSET_UPPER_NUMERIC
        rf.valid_name_length = 8
        rf.valid_power_levels = list(reversed(POWER_LEVELS))
        rf.valid_skips = ['', 'S']
        return rf

    def _get_zone_info(self):
        zones = {}
        for i in range(128):
            info = self._memobj.zoneinfo[i]
            if info.used:
                zones[i + 1] = int(info.count)
        if not zones:
            LOG.warning('No zones defined, defaulting to 1')
            # Default to at least one zone so we have *something*
            zones[1] = 0
        return zones

    def _set_num_zones(self, count):
        LOG.debug('Setting zones to %i' % count)
        for i in range(128):
            index_obj = self._memobj.index[i]
            if index_obj.zone != 0xFF and index_obj.zone > count:
                LOG.debug('Deleting memory %i:%i to delete zone',
                          index_obj.zone, index_obj.memory)
                self._delete_entry_from_index(i)

        # This will disable zones with no memories, so do it after deletion
        # but before we may expand (if we're increasing)
        self.recalculate_index_entry_bitmask()

        unused_zones = []
        for i in range(128):
            info = self._memobj.zoneinfo[i]
            info.used = i < count
            if not info.used:
                unused_zones.append(i + 1)

    def _get_zone(self, zone):
        class TKx160Zone(self.__class__):
            MODEL = self.MODEL
            VARIANT = 'Zone %i' % zone
            VALID_BANDS = self.VALID_BANDS

            def __init__(self, parent, zone):
                self._memobj = parent._memobj
                self._zone = zone

        return TKx160Zone(self, zone)

    def get_sub_devices(self):
        if not self._memobj:
            return [self._get_zone(1)]
        return [self._get_zone(zone)
                for zone, count in self._get_zone_info().items()]

    def get_raw_memory(self, number):
        index_entry, slot = self.check_index(number)
        return (repr(self._memobj.index[index_entry]) +
                repr(self._memobj.memories[slot]))

    def check_index(self, number):
        """Return the index entry and slot for a channel in our zone"""
        for i in range(128):
            index_obj = self._memobj.index[i]
            if index_obj.zone == self._zone and index_obj.memory == number:
                LOG.debug('Found %i:%i in slot %i',
                          self._zone, number,
                          index_obj.slot)
                return i, index_obj.slot
        return None, None

    def next_slot(self, number):
        """Claim the next available slot entry for channel."""
        for next_index in range(128):
            index_obj = self._memobj.index[next_index]
            if index_obj.zone == 0xFF:
                break
        else:
            raise errors.RadioError('Radio is out if index memory')

        for next_slot in range(128):
            if self._memobj.memories[next_slot].number == 0xFF:
                break
        else:
            raise errors.RadioError('Radio is out of memory slots')

        LOG.debug('Putting channel %i in index spot %i',
                  number, next_index)
        index_obj = self._memobj.index[next_index]
        index_obj.zone = self._zone
        index_obj.memory = number
        index_obj.zero = 0
        index_obj.slot = next_slot

        self.recalculate_index_entry_bitmask()

        return next_slot

    def _delete_entry_from_index(self, index_entry):
        last_entry = index_entry

        # Find the last index entry
        for i in reversed(range(0, 128)):
            mask = 1 << (i % 8)
            if self._memobj.flag1[i // 8] & mask:
                last_entry = i
                break

        src_obj = self._memobj.index[last_entry]
        dst_obj = self._memobj.index[index_entry]
        if index_entry != last_entry:
            LOG.debug('Moving index entry %i to %i for delete of %i:%i',
                      last_entry, index_entry,
                      dst_obj.zone, dst_obj.memory)
            # Move last to current
            dst_obj.zone = src_obj.zone
            dst_obj.memory = src_obj.memory
            dst_obj.zero = src_obj.zero
            dst_obj.slot = src_obj.slot

        # Clear out the entry we moved, which may be the last and only one
        src_obj.set_raw(b'\xFF' * 4)

    def delete_from_index(self, number):
        """Delete the index entry for number"""
        # Find the index entry to delete
        index_entry, _slot = self.check_index(number)
        self._delete_entry_from_index(index_entry)
        self.recalculate_index_entry_bitmask()

    def recalculate_index_entry_bitmask(self):
        # Recalculate the pins. This is more laborious than needed, but should
        # also clean up residue.
        zone_counts = [0] * 128
        mem_flags = [0] * 16
        for i in range(128):
            index_obj = self._memobj.index[i]
            byte = i // 8
            mask = 1 << (i % 8)
            if index_obj.memory != 0xFF:
                mem_flags[byte] |= mask
                zone_counts[index_obj.zone - 1] += 1

        for i, byte in enumerate(mem_flags):
            self._memobj.flag1[i] = byte
            self._memobj.flag2[i] = byte

        zone_flags = [0] * 16
        # Update zoneinfo
        for i, count in enumerate(zone_counts):
            self._memobj.zoneinfo[i].used = bool(count)
            self._memobj.zoneinfo[i].count = count
            if count:
                zone_flags[i // 8] |= (1 << (i % 8))
                LOG.debug('Zone %i count is %i', i + 1, count)

        for i, byte in enumerate(zone_flags):
            self._memobj.zoneflag[i] = byte

    def get_memory(self, number):
        m = chirp_common.Memory(number)
        _index_entry, slot = self.check_index(number)
        if slot is None:
            m.empty = True
            return m

        _mem = self._memobj.memories[slot]

        m.freq = int(_mem.rx_freq) * 10
        offset = int(_mem.tx_freq) * 10 - m.freq
        if _mem.tx_freq.get_raw() == b'\xFF\xFF\xFF\xFF':
            m.offset = 0
            m.duplex = 'off'
        elif offset < 0:
            m.offset = abs(offset)
            m.duplex = '-'
        elif offset > 0:
            m.offset = offset
            m.duplex = '+'
        else:
            m.offset = 0

        rxtone = tk8180.KenwoodTKx180Radio._decode_tone(
            _mem.rxtone)
        txtone = tk8180.KenwoodTKx180Radio._decode_tone(
            _mem.txtone)
        chirp_common.split_tone_decode(m, txtone, rxtone)

        m.name = str(_mem.name).rstrip()
        m.power = POWER_LEVELS[int(_mem.highpower)]
        m.mode = 'FM' if _mem.wide else 'NFM'
        m.skip = '' if _mem.scanadd else 'S'

        return m

    def set_memory(self, mem):
        _index_entry, slot = self.check_index(mem.number)
        if mem.empty:
            if slot is not None:
                self._memobj.memories[slot].set_raw(b'\xFF' * 32)
                self.delete_from_index(mem.number)
            return
        elif slot is None:
            slot = self.next_slot(mem.number)

        _mem = self._memobj.memories[slot]
        _mem.number = mem.number
        _mem.zone = self._zone

        LOG.debug('Setting memory %i in slot %i', mem.number, slot)

        _mem.rx_freq = mem.freq // 10
        if mem.duplex == '':
            _mem.tx_freq = mem.freq // 10
        elif mem.duplex == 'split':
            _mem.tx_freq = mem.offset // 10
        elif mem.duplex == 'off':
            _mem.tx_freq.set_raw(b'\xFF' * 4)
        elif mem.duplex == '-':
            _mem.tx_freq = (mem.freq - mem.offset) // 10
        elif mem.duplex == '+':
            _mem.tx_freq = (mem.freq + mem.offset) // 10
        else:
            raise errors.RadioError('Unsupported duplex mode %r' % mem.duplex)

        txtone, rxtone = chirp_common.split_tone_encode(mem)
        _mem.rxtone = tk8180.KenwoodTKx180Radio._encode_tone(*rxtone)
        _mem.txtone = tk8180.KenwoodTKx180Radio._encode_tone(*txtone)

        _mem.wide = mem.mode == 'FM'
        _mem.highpower = mem.power == POWER_LEVELS[1]
        _mem.scanadd = mem.skip == ''
        _mem.name = mem.name[:8].ljust(8)

        # Set the flags we don't support
        _mem.bcl = 0
        _mem.beatshift = 0
        _mem.pttid = 0
        _mem.compander = 0
        _mem.pttidmute = 0
        _mem.optsig = 0

        # Set the unknowns
        _mem.unknown1[0] = 0x02
        _mem.unknown1[1] = 0x02
        _mem.unknown2_0 = 0x0F
        _mem.unknown2_01 = 0x00
        _mem.unknown2_1 = 0x00
        _mem.unknown2_2 = 0x00
        _mem.unknown2_3 = 0x00
        _mem.unknown3_0 = 0x00
        _mem.unknown3_1 = 0x00
        _mem.unknown4 = 0x1F
        _mem.unknown5[0] = 0xFF
        _mem.unknown5[1] = 0xFF

    @staticmethod
    def make_key_group(keysobj, keynames, keyvals):
        keyvals_by_label = {v: k for k, v in keyvals.items()}

        def applycb(setting):
            element = setting.get_name()
            value = keyvals_by_label[str(setting.value)]
            LOG.debug('Setting %s=%s (0x%02x)', element, setting.value, value)
            setattr(keysobj, element, value)

        group = settings.RadioSettingGroup('keys', 'Keys')
        for name, element in keynames.items():
            current = keyvals[int(getattr(keysobj, element))]
            val = settings.RadioSettingValueList(keyvals.values(), current)
            keysetting = settings.RadioSetting(element, name, val)
            keysetting.set_apply_callback(applycb)
            group.append(keysetting)
        return group

    def get_settings(self):
        num_zones = len(self._get_zone_info())
        zones = settings.RadioSettingGroup('zones', 'Zones')
        zone_count = settings.RadioSetting(
            'zonecount', 'Number of Zones',
            settings.RadioSettingValueInteger(1, 128, num_zones))
        zone_count.set_doc('Number of zones in the radio. '
                           'Requires a save and re-load of the file to take '
                           'effect. Reducing this number will DELETE '
                           'memories in the affected zones!')
        zones.append(zone_count)
        keys = self.make_key_group(self._memobj.keys,
                                   KEY_NAMES,
                                   KEYS)
        return settings.RadioSettings(zones, keys)

    def set_settings(self, _settings):
        for element in _settings:
            if not isinstance(element, settings.RadioSetting):
                self.set_settings(element)
            elif element.get_name() == 'zonecount':
                self._set_num_zones(int(element.value))

    def save_mmap(self, filename):
        if filename.lower().endswith('.dat'):
            with open(filename, 'wb') as f:
                f.write(b'EX3774'.ljust(16, b'\xFF'))
                f.write(self._mmap.get_packed())
                LOG.info('Wrote DAT file')
        else:
            super().save_mmap(filename)

    def load_mmap(self, filename):
        if filename.lower().endswith('.dat'):
            with open(filename, 'rb') as f:
                header = f.read(16)
                LOG.debug('DAT header:\n%s' % util.hexprint(header))
                self._mmap = memmap.MemoryMapBytes(f.read())
                LOG.info('Loaded DAT file')
            self.process_mmap()
        else:
            super().load_mmap(filename)


@directory.register
class TK7160RadioM(TKx160Radio):
    MODEL = 'TK-7160M'
    VALID_BANDS = [(136000000, 174000000)]
    _model = b'M66- \x04'


@directory.register
class TK7160RadioK(TKx160Radio):
    MODEL = 'TK-7160K'
    VALID_BANDS = [(136000000, 174000000)]
    _model = b'M66- \x02'


@directory.register
class TK8160RadioK(TKx160Radio):
    MODEL = 'TK-8160K'
    VALID_BANDS = [(400000000, 520000000)]
    _model = b'M67- "'


@directory.register
class TK8160RadioM(TKx160Radio):
    MODEL = 'TK-8160M'
    VALID_BANDS = [(400000000, 520000000)]
    _model = b'M67- D'
